
import numpy as np
from ..qt_compat import QtGui, QtCore
from collections import OrderedDict

import logging as log

from .model import DataModel
from .launcher import Launcher
from .singleton import Singleton

Axial = DataModel.instance().Axial
Sagittal = DataModel.instance().Sagittal
Coronal = DataModel.instance().Coronal


class Label(object):

    def __init__(self, name, idx, color='#000000', visible=True,
                 parent_level=-1, parent_label=-1):
        self.name = name
        self.idx = idx
        self.color = color
        self.visible = visible
        self.parent_level = parent_level
        self.parent_label = parent_label

@Singleton
class LabelManager(QtCore.QObject):

    levelAdded = QtCore.pyqtSignal(int, str)
    levelLoaded = QtCore.pyqtSignal(int, str)
    levelRemoved = QtCore.pyqtSignal(int, str, bool)
    saveLevel = QtCore.pyqtSignal(int, str)

    labelAdded = QtCore.pyqtSignal(int, str, int, str)
    labelLoaded = QtCore.pyqtSignal(int, str, int, str, str, bool, int, int)
    labelRemoved = QtCore.pyqtSignal(int, str, object)
    labelUpdated = QtCore.pyqtSignal(int, str)
    labelSelected = QtCore.pyqtSignal(int, str, int)
    labelNameChanged = QtCore.pyqtSignal(int, str, int, str)
    labelColorChanged = QtCore.pyqtSignal(int, str, int, str)
    labelVisibilityChanged = QtCore.pyqtSignal(int, str, int, bool)
    labelParentChanged = QtCore.pyqtSignal(int, str, int, int, int)

    def __init__(self):
        QtCore.QObject.__init__(self)
        self.DM = DataModel.instance()
        self._levels = OrderedDict()
        self._counts = OrderedDict()
        self._datasets = OrderedDict()
        self.next_level = 0

    def len(self, level=None):
        if level is None:
            return len(self._levels)
        else:
            return len(self._levels[level])

    def levels(self):
        return list(self._levels.keys())

    def labels(self, level):
        return list(self._levels[level].values())

    def dataset(self, level):
        return self._datasets[level]

    def idxs(self, level):
        return [label.idx for label in self._levels[level].values()]

    def colors(self, level):
        return [label.color for label in self._levels[level].values()]

    def names(self, level):
        return [label.name for label in self._levels[level].values()]

    def parent_levels(self, level):
        return [label.parent_level for label in self._levels[level].values()]

    def parent_labels(self, level):
        return [label.parent_label for label in self._levels[level].values()]

    def visibility(self, level):
        return [label.visible for label in self._levels[level].values()]

    def foundLevel(self, levelidx):
        if levelidx >= self.next_level:
            self.next_level = levelidx + 1

    def loadLevel(self, levelidx, dataset):
        if levelidx in self._levels:
            log.error('  --- Annotations [{}] has same ID than [{}]. Skipping.'.format(
                dataset, self.dataset(levelidx)
            ))
            self.DM.set_attrs(dataset, dict(active=False))
            return
        self._levels[levelidx] = OrderedDict()
        self._counts[levelidx] = 0
        self._datasets[levelidx] = dataset
        self.levelLoaded.emit(levelidx, dataset)
        if levelidx >= self.next_level:
            self.next_level = levelidx + 1

        attrs = self.DM.attrs(dataset)
        label = attrs['label']
        names = attrs['names']
        colors = attrs['colors']
        visible = attrs['visible']
        parent_levels = attrs['parent_levels']
        parent_labels = attrs['parent_labels']

        if label is not None and len(label) > 0:
            for l, n, c, v, ple, pla in zip(label, names, colors, visible,
                                            parent_levels, parent_labels):
                self.loadLabel(levelidx, l, n, c, v > 0, ple, pla)

        return levelidx, dataset

    def addLevel(self):
        levelidx = self.next_level
        dataset = 'annotations/annotations{}'.format(levelidx)
        self._levels[levelidx] = OrderedDict()
        self._counts[levelidx] = 0
        self._datasets[levelidx] = dataset
        self.levelAdded.emit(levelidx, dataset)
        self.next_level += 1
        return levelidx, dataset

    def removeLevel(self, level, force=False):
        dataset = self._datasets[level]
        del self._levels[level]
        del self._counts[level]
        del self._datasets[level]
        self.levelRemoved.emit(level, dataset, force)

    def addLabel(self, level):
        idx = self._counts[level]
        name = 'Label {}'.format(idx)
        self._levels[level][idx] = Label(name, idx)
        self.labelAdded.emit(level, self._datasets[level], idx, name)
        self._counts[level] += 1

    def loadLabel(self, level, label, name, color, visible, parent_level, parent_label):

        # Fix for Python3 string issues
        try:
            name = name.decode('UTF-8')
            color = color.decode('UTF-8')
        except AttributeError as e:
            pass

        self._levels[level][label] = Label(name, label, color, visible, parent_level, parent_label)

        self.labelLoaded.emit(level, self._datasets[level], label, name, color,
                              visible, parent_level, parent_label)
        if label >= self._counts[level]:
            self._counts[level] = label+1

    def get(self, level, label):
        """
        Retrieve a Label from the _levels dict. If not there, return an empty Label object.
        :param level: Level index  
        :param label: Label index
        :return: A Label object
        """
        return self._levels.get(level, {}).get(label, Label('Empty', -1))

    def removeLabel(self, level, label):
        if label in self._levels[level]:
            labelobj = self._levels[level][label]
            del self._levels[level][label]
            self.labelRemoved.emit(level, self._datasets[level], labelobj)

    def isin(self, level, label):
        return label in self._levels[level]

    def index(self, level, label):
        return list(self._levels[level].keys()).index(name)

    def clear(self, level):
        self._levels[level].clear()

    def setLabelParent(self, level, label, parent_level, parent_label):
        self._levels[level][label].parent_level = parent_level
        self._levels[level][label].parent_label = parent_label
        self.labelParentChanged.emit(level, self._datasets[level], label,
                                     parent_level, parent_label)

    def changeLabelName(self, level, label, name):
        self._levels[level][label].name = name
        self.labelNameChanged.emit(level, self._datasets[level], label, name)

    def changeLabelColor(self, level, label, color):
        self._levels[level][label].color = color
        self.labelColorChanged.emit(level, self._datasets[level], label, color)

    def changeLabelVisibility(self, level, label, visible):
        self._levels[level][label].visible = visible
        self.labelVisibilityChanged.emit(level, self._datasets[level], label, visible)

    def selectLabel(self, level, label):
        self.labelSelected.emit(level, self._datasets[level], label)

    def save(self, level):
        self.saveLevel.emit(level, self._datasets[level])

    def saveAll(self):
        for level in self.levels():
            self.saveLevel.emit(level, self._datasets[level])




from ..qt_compat import QtGui, QtCore

import numpy as np
import logging as log
from collections import deque

import os
import ast
import glob
import h5py as h5
from pickle import dumps, loads

from scipy.stats import gaussian_kde

from .singleton import Singleton


@Singleton
class DataModel(QtCore.QObject):

    roi_changed = QtCore.pyqtSignal()
    select_channel = QtCore.pyqtSignal(str)
    update_channel = QtCore.pyqtSignal(str)
    cropped = QtCore.pyqtSignal()
    vmin_changed = QtCore.pyqtSignal(float)
    vmax_changed = QtCore.pyqtSignal(float)
    evmin_changed = QtCore.pyqtSignal(float)
    evmax_changed = QtCore.pyqtSignal(float)
    channel_computed = QtCore.pyqtSignal(str, dict)
    channel_removed = QtCore.pyqtSignal(str)
    feature_updated = QtCore.pyqtSignal(str)
    classifier_trained = QtCore.pyqtSignal(int)
    voxel_descriptor_computed = QtCore.pyqtSignal()
    voxel_descriptor_removed = QtCore.pyqtSignal()
    supervoxel_descriptor_computed = QtCore.pyqtSignal()
    supervoxel_descriptor_removed = QtCore.pyqtSignal()
    supervoxel_descriptor_cleared = QtCore.pyqtSignal()
    level_predicted = QtCore.pyqtSignal(int)
    clf_channel_computed = QtCore.pyqtSignal(int, str, str, bool, dict)

    Axial = 0
    Sagittal = 1
    Coronal = 2

    def __init__(self):
        QtCore.QObject.__init__(self)

        self.main_window = None

        # tmp_data
        self.tmp_data = None
        self.tmp_stats = None

        # Storage
        self.wspath = None
        self.current_idx = 0
        self.selected_gpu = -1

        # DataFiles
        self.data = None
        self.data_shape = None
        self.active_roi = slice(None), slice(None), slice(None)

        # Superpixels
        self.svlabels = None
        self.svindex = None
        self.svtable = None
        self.svtotal = None

        # Megavoxels
        self.mvlabels = None
        self.mvindex = None
        self.mvtable = None
        self.mvtotal = None

        # Classifier
        self.clf = None
        self.clf_name = None
        self.clf_saved_name = None
        self.clf_params = None
        self.X_train = None
        self.y_train = None

        # Parameters to be saved with classifier
        self.desc_params = None

        # GT Labels
        self.gtlevel = 0
        self.gtradius = 1
        self.gtinterpolation = 'linear'
        self.gtselected = None
        self.last_changes = deque(maxlen=10)
        self.redo_changes = deque(maxlen=10)
        self.growing_bbox = ['data', 1., [10, 50, 50], 0]

        # KDE plots
        self.computed_histograms = dict()
        self.channel_computed.connect(self.disable_histogram)

    def disable_histogram(self, name):
        if name in self.computed_histograms:
            del self.computed_kde[name]

    def get_histogram(self, dataset):
        if dataset in self.computed_histograms:
            return self.computed_histograms[dataset]

        log.info('+ Loading channel into memory')
        data = self.load_ds(dataset).ravel()
        attr = self.attrs(dataset)
        amin = attr['vmin']
        amax = attr['vmax']

        log.info('+ Computing Histogram')
        y, x = np.histogram(data[~np.isnan(data)], 10000)
        x = (x[:-1] + x[1:]) / 2.

        self.computed_histograms[dataset] = (x, y)
        return (x, y)

    ##########################################################################
    # TOI Transformation
    ##########################################################################

    def region_shape(self):
        return tuple(map(lambda x: x.stop - x.start, self.active_roi))

    def transform_slices(self, slice_z=None, slice_y=None, slice_x=None):
        slices = []

        for t, s in zip(self.active_roi, (slice_z, slice_y, slice_x)):
            start = t.start
            stop = t.stop

            try:
                s = int(s)
                start += s
                stop = start + 1
            except:
                if s is None:
                    s = slice(None)
                if s.stop is not None:
                    stop = t.start + s.stop
                if s.start is not None:
                    start += s.start

            slices.append(slice(start, stop))

        return slices

    ##########################################################################
    # DATASETS IN / OUT
    ##########################################################################

    def create_empty_dataset(self, dataset, shape, dtype, check=True,
                             params=None, fillvalue=-1):
        ds_file = self.ds_path(dataset)
        ds_folder = os.path.dirname(ds_file)
        if not os.path.exists(ds_folder):
            os.mkdir(ds_folder)
        if check and os.path.isfile(ds_file):
            return False
        with h5.File(ds_file, 'w') as f:
            ds = f.create_dataset('data', shape=shape, fillvalue=fillvalue,
                                  dtype=dtype)
            if params is not None:
                for k, v in params.items():
                    ds.attrs[k] = v
        return True

    def write_dataset(self, dataset, data, params=None):
        ds_file = self.ds_path(dataset)
        ds_folder = os.path.dirname(ds_file)
        if not os.path.exists(ds_folder):
            os.mkdir(ds_folder)
        with h5.File(ds_file, 'a') as f:
            if 'data' in f:
                f['data'].write_direct(data)
            else:
                f.create_dataset('data', data=data)
            if params is not None:
                for k, v in params.items():
                    f['data'].attrs[k] = v

    def load_ds(self, dataset):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            data = np.zeros_like(f['data'])
            f['data'].read_direct(data)
        return data

    def load_slices(self, dataset, slice_z=None, slice_y=None, slice_x=None,
                    apply_roi=True):
        ds_file = self.ds_path(dataset)
        if apply_roi:
            sz, sy, sx = self.transform_slices(slice_z, slice_y, slice_x)
        else:
            sz, sy, sx = slice_z, slice_y, slice_x
        with h5.File(ds_file, 'r') as f:
            data = f['data'][sz, sy, sx]
        return data

    def write_slices(self, dataset, data, slice_z=None, slice_y=None,
                     slice_x=None, params=None, apply_roi=True):
        ds_file = self.ds_path(dataset)
        if apply_roi:
            sz, sy, sx = self.transform_slices(slice_z, slice_y, slice_x)
        else:
            sz, sy, sx = slice_z, slice_y, slice_x
        with h5.File(ds_file, 'a') as f:
            f['data'][sz, sy, sx] = data
            if params is not None:
                for k, v in params.items():
                    f['data'].attrs[k] = v
        return data

    def remove_dataset(self, dataset):
        ds_file = self.ds_path(dataset)
        if os.path.exists(ds_file):
            os.remove(ds_file)

    def add_training_data(self, X_train, y_train):
        """
        Adds data used to train classifier to the data model
        :param X_train: Values
        :param y_train: Labels
        """
        self.X_train = X_train
        self.y_train = y_train

    def has_training_data(self):
        return self.X_train is not None and self.y_train is not None

    def get_training_data(self):
        """
        :return: A tuple with the training data and labels
        """
        if self.has_training_data():
            return self.X_train, self.y_train

    def new_annotations_added(self, y_data):
        data = self.load_slices(y_data)
        return np.any(data > -1)

    def save_classifier(self, path, lbl_attrs):
        """
        Save classifier to filepath
        :param path: filepath for classifier output
        """
        if self.has_classifier():
            clf_pkl = dumps(self.clf)
            with h5.File(path, 'w') as out_file:
                if self.has_training_data():
                    out_file['X_train'] = self.X_train
                    out_file['y_train'] = self.y_train
                out_file['classifier'] = np.void(clf_pkl)
                out_file['classifier'].attrs['class'] = type(self.clf).__name__
                out_file['classifier'].attrs['features'] = self.desc_params.get('features')
                out_file['classifier'].attrs['parameters'] = str(self.clf_params)
                # Add the label information
                self.write_attrs_to_empty_dataset(out_file, "lbl_info", lbl_attrs)
                self.save_supervox_and_filter_settings(out_file, has_classifier=True)
        else:
            log.error("Data Model has no classifier!")

    def save_settings_file(self, path):
        """
        Saves a supervoxel and filter settings HDF5 file
        :param path: File path to save to
        """
        with h5.File(path, 'w') as out_file:
            self.save_supervox_and_filter_settings(out_file, has_classifier=False)

    def save_supervox_and_filter_settings(self, out_file, has_classifier):
        # Copy the supervoxel information across
        svds = 'supervoxels/supervoxels'
        supervox_attrs = self.attrs(svds) # returns None if supervoxels don't exist
        if supervox_attrs:
            self.write_attrs_to_empty_dataset(out_file, "supervox_info", supervox_attrs)
        # Get the metadata for the feature channels
        result_list = self.get_channel_metadata(has_classifier)
        # Fix for case where source feature for supervoxel calculation is not used for prediction
        if supervox_attrs:
            sv_source = supervox_attrs['source']
            sv_source_name = sv_source.split('/')[1]
            sv_source_index = sv_source_name.split('_')[0]
            # Check if the supervoxel source is in the result_list
            if not any(d.get('feature_idx', None) == int(sv_source_index) for d in result_list):
                # It is not there, add it from the channel metadata
                sv_source_attrs = self.attrs(sv_source)
                sv_source_attrs['unique_sv_source'] = True
                result_list.append(sv_source_attrs)

        # Create the datasets
        channel_dict = {}
        for result in result_list:
            dataset_name = "{}_{}".format(result.get('feature_idx', ''), result.get('feature_type', 'data'))
            channel_dict.setdefault('channel_list', []).append(dataset_name)
            # fix cases where data is used as an input..
            if dataset_name == "_data":
                dataset_name = dataset_name[1:]
            self.write_attrs_to_empty_dataset(out_file, dataset_name, result)
        # Write a dataset with the channel list as an attribute
        self.write_attrs_to_empty_dataset(out_file, "channel_list", channel_dict)

    def write_attrs_to_empty_dataset(self, file, dataset, attrs_dict):
        """
        Helper function to create and empty dataset in a file and 
        populate HDF5 dataset attrs
        :param file: H5py file object
        :param dataset: Name of the dataset to create (str)
        :param attrs_dict: Dictionary of attributes to add to dataset
        """
        file.create_dataset(dataset, dtype='f')
        for k, v in attrs_dict.items():
            file[dataset].attrs[k] = v

    def load_saved_settings_from_file(self, path, has_classifier=False):
        """
        Returns a dictionary containing all the metadata for settings saved
        :param path: Path to classifier or settings HDF5 file
        :return: Dictionary with metadata
        """
        result = {}
        if os.path.exists(path):
            with h5.File(path, 'r') as in_file:
                try:
                    if has_classifier:
                        # Get the list of channels used by the classifier
                        channel_list = in_file['classifier'].attrs['features']
                        # Store dicts with the metadata for each channel in the list
                        for channel_string in channel_list:
                            self.add_channel_to_result(channel_string, in_file, result)
                        result['lbl_attrs'] = dict(in_file['lbl_info'].attrs.items())
                    else:
                        # This is a settings file
                        channel_list = in_file['channel_list'].attrs['channel_list']
                        result['channel_list'] = channel_list
                        for channel in channel_list:
                            result[channel] = dict(in_file[channel].attrs)
                            result[channel]['out'] = "channels/{}".format(channel)

                    sv_attrs = in_file['supervox_info'].attrs
                    if sv_attrs:
                        result['sv_attrs'] = dict(sv_attrs.items())

                        # Check if the source channel for calculating supervoxels is present in channel list, add if not
                        sv_source = result['sv_attrs']['source']
                        name = sv_source.split('/')[1]
                        if name not in result['channel_list']:
                            log.info("Supervoxel source channel is not a feature channel. Adding to result dictionary.")
                            self.add_channel_to_result(sv_source, in_file, result)
                except Exception as e:
                    log.error("Could not load attributes: {}".format(e))
            if "data" in result:  # Special case - data has been used for prediction
                result['data']['feature_name'] = "data"
                result['data']['feature_-dx'] = -1
            return result
        else:
            log.error("File does not appear to exist")

    def add_channel_to_result(self, channel_string, in_file, result):
        # channel_string is in form 'channel/name'
        name = channel_string.split('/')[1]
        result[name] = dict(in_file[name].attrs)
        result.setdefault('channel_list', []).append(name)
        result[name]['out'] = channel_string

    def get_classifier_params(self):
        return self.clf_params

    def get_channel_metadata(self, classifier=True):
        """
        Get the required metadata from each feature channel file
        :return metadata: list of dictionaries, each containing data for one channel
        """
        if classifier:
            return [self.attrs(feature) for feature in self.desc_params.get('features')]
        results_list = []
        for channel in self.available_channels():
            results_list.append(self.attrs(channel[1]))
        return results_list

    def add_classifier_to_model(self, clf, clf_params):
        """
        Add classifier to model object along with parameters
        :param clf: classifier to add to object
        """
        self.clf = clf
        self.clf_name = clf.__class__.__name__
        self.clf_params = clf_params


    def get_classifier_from_model(self):
        """
        Retrieves classifier from model
        :return: classifier
        """
        if not self.clf:
            log.error("No Classifier in model")
        else:
            return self.clf

    def has_classifier(self):
        return self.clf is not None

    def load_classifier(self, path):
        if os.path.exists(path):
            with h5.File(path, 'r') as in_file:
                try:
                    clf_str = in_file['classifier']
                    clf_params = ast.literal_eval(in_file['classifier'].attrs['parameters'])
                    clf = loads(clf_str[()].tostring())
                    class_str =  in_file['classifier'].attrs['class']
                    X_train = in_file["X_train"][()]
                    y_train = in_file["y_train"][()]
                    self.add_training_data(X_train, y_train)
                except Exception as e:
                    log.error(path, e)
                    return False

            if str(class_str) == type(clf).__name__:
                self.add_classifier_to_model(clf, clf_params)
                path = os.path.normpath(path)
                split_path = path.split(os.sep)
                self.clf_saved_name = " - ".join(split_path[-3::2])
                log.info("Classifier {} Loaded: {}".format(self.clf_saved_name, type(self.clf)))
                return True
            else:
                log.error("Classifier not loaded. Class description {} does not"
                          " match classifier type {}".format(str(class_str), type(clf).__name__))
                return False
        else:
            log.error("File does not appear to exist")
            return False

    def add_desc_params_to_model(self, desc_params):
        """
        Add desc_params to model object
        :param desc_params: dictionary of descriptors
        """
        self.desc_params = desc_params

    ##########################################################################
    # ATTRIBUTES AND MEMBERSHIP
    ##########################################################################

    def set_attrs(self, dataset, attrs):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'a') as f:
            for k, v in attrs.items():
                if isinstance(v, list): #this helps with python 3 compatibility
                    if len(v)!=0:
                        if isinstance(v[0], str) or isinstance(v[0], bytes):
                            v = [np.string_(x) for x in v]
                f['data'].attrs[k] = v

    def attrs(self, dataset):
        ds_file = self.ds_path(dataset)
        try:
            with h5.File(ds_file, 'r') as f:
                attr = dict(f['data'].attrs)
            return attr
        except OSError: # File doesn't exist
            return None

    def attr(self, dataset, attr, value=None):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            if attr in f['data'].attrs:
                if value is None:
                    val = f['data'].attrs[attr]
                else:
                    val = (f['data'].attrs[attr] == value)
            else:
                val = None
        return val

    def ds_shape(self, dataset):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            shape = f['data'].shape
        return shape

    def has_ds(self, dataset):
        active = False
        ds_file = self.ds_path(dataset)
        if os.path.isfile(ds_file):
            with h5.File(ds_file, 'r') as f:
                if 'data' in f and 'active' in f['data'].attrs:
                    active = f['data'].attrs['active']
        return active

    def has_grp(self, group):
        ds_folder = self.grp_path(group)
        return os.path.exists(ds_folder)

    def ds_path(self, dataset):
        if dataset.startswith('/'):
            dataset = dataset[1:]
        return os.path.join(self.wspath, '{}.h5'.format(dataset))

    def grp_path(self, group):
        return os.path.join(self.wspath, group)

    ##########################################################################
    # SCAN UTILITIES
    ##########################################################################

    def scan_datasets_group(self, group, shape=None, dtype=None, path=''):
        datasets = []
        for name, ds in group.items():
            curr_path = '{}/{}'.format(path, name)
            if hasattr(ds, 'shape'):
                if len(ds.shape) == 3 \
                        and (shape is None or ds.shape == shape) \
                        and (dtype is None or ds.dtype == dtype):
                    datasets.append(curr_path)
            else:
                extra = self.scan_datasets_group(ds, shape=shape, path=curr_path)
                if len(extra) > 0:
                    datasets += extra
        return datasets


    def available_hdf5_datasets(self, path, shape=None, dtype=None):
        datasets = []
        with h5.File(path, 'r') as f:
            datasets = self.scan_datasets_group(f, shape=shape, dtype=dtype)
        return datasets

    ##########################################################################
    # FEATURES AND DESCRIPTORS
    ##########################################################################

    def find_dataset(self, group, filter_active=True):
        grp_dir = self.grp_path(group)
        ds = glob.glob('{}/*.h5'.format(grp_dir))
        ds = ['{}/{}'.format(group, os.path.basename(d[:-3])) for d in ds]
        if filter_active:
            ds = [d for d in ds if self.attr(d, 'active') == True]
        return ds

    def available_channels(self, return_names=False, filter_active=True):
        if self.has_grp('channels'):
            possible = self.find_dataset('channels', filter_active=filter_active)
            af = []
            for f in possible:
                attrs = self.attrs(f)
                if 'active' in attrs and 'feature_idx' in attrs:
                    fidx = attrs['feature_idx']
                    if return_names:
                        af.append((fidx, f, attrs))
                    else:
                        af.append((fidx, f))
        else:
            af = []
        return af

    def available_annotations(self, return_names=False):
        if self.has_grp('annotations'):
            possible = self.find_dataset('annotations')
            af = []
            for f in possible:
                attrs = self.attrs(f)
                if 'active' in attrs and 'levelid' in attrs:
                    af.append((attrs['levelid'], f, attrs['active']))
        else:
            af = []
        return af





import h5py as h5
import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

import os
import logging as log

try:
    from tifffile import imsave as tiffsave
    tiff_enabled = True
except:
    tiff_enabled = False

from ..lib.io import MRC
from ..core import DataModel, LabelManager, Launcher
from ..widgets import HeaderLabel, RoundedWidget, TComboBox, HWidgets, \
                      SubHeaderLabel, ActionButton, FileWidget, SourceCombo

from .. import actions as ac
from .base import Plugin

from skimage.util import img_as_ubyte


class CheckableLevel(RoundedWidget):

    def __init__(self, level_id, level_name, parent=None):
        super(CheckableLevel, self).__init__(parent=parent, color=None,
                                         bg='#cde5e5', width=0)
        self.idx = level_id
        self.name = level_name

        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(hbox)

        self.select = QtWidgets.QCheckBox()
        hbox.addWidget(self.select)

        self.name = QtWidgets.QLabel(level_name)
        self.name.setStyleSheet('font-size: 12pt; color: #009999;'
                                'font-weight: bold;')
        hbox.addWidget(self.name, 1)

        self.isChecked = self.select.isChecked
        self.setChecked = self.select.setChecked

    def mousePressEvent(self, ev):
        self.select.setChecked(not self.select.isChecked())

    def value(self):
        return self.idx


class Export(Plugin):

    name = 'Export'

    def __init__(self):
        super(Export, self).__init__(ptype=Plugin.Plugin)

        self.DM = DataModel.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        self.addWidget(HeaderLabel('Export Segmentations'))
        dummy = QtWidgets.QWidget()
        vbox = QtWidgets.QVBoxLayout()
        dummy.setLayout(vbox)
        self.addWidget(dummy)

        vbox.addWidget(SubHeaderLabel('Available Annotation Levels'))
        self.levels = {}
        self.container = QtWidgets.QWidget()
        self.container.setLayout(QtWidgets.QVBoxLayout())
        vbox.addWidget(self.container)

        vbox.addWidget(SubHeaderLabel('Export Levels'))
        self.output = FileWidget(folder=True)
        vbox.addWidget(HWidgets('Output Folder:', self.output, stretch=[0,1]))

        options = ['Raw Data', 'Raw Annotations', 'Segmentation Masks', 'Masked Data']
        self.combo = TComboBox('Output:', options, selected=0)
        self.combo.currentIndexChanged.connect(self.on_combo)
        vbox.addWidget(HWidgets(self.combo, None, stretch=[0,1]))

        self.source = SourceCombo()
        self.scale = QtWidgets.QCheckBox('Normalize')
        self.scale.setChecked(True)
        self.invert = QtWidgets.QCheckBox('Invert')
        self.scont = HWidgets(self.source, self.scale, self.invert, stretch=[1,0,0])
        self.scont.setVisible(False)
        vbox.addWidget(self.scont)

        formats = ['HDF5 (.h5)', 'MRC (.mrc)', 'Tiff Stack (.tiff)']
        self.format = TComboBox('Format:', formats, selected=0)
        self.overwrite = QtWidgets.QCheckBox('Overwrite')
        self.overwrite.setChecked(True)
        self.export = ActionButton('Export')
        vbox.addWidget(HWidgets(self.format, self.overwrite, self.export,
                                stretch=[1,0,0]))

        self.LBLM.levelAdded.connect(self.on_level_added)
        self.LBLM.levelLoaded.connect(self.on_level_added)
        self.LBLM.levelRemoved.connect(self.on_level_removed)
        self.export.clicked.connect(self.on_export)

    def on_combo(self, idx):
        self.scont.setVisible(idx == 3)

    def on_level_added(self, level, dataset):
        if level not in self.levels:
            obj = CheckableLevel(level, 'Level {}'.format(level))
            self.levels[level] = obj
            self.container.layout().addWidget(obj)

    def on_level_removed(self, level, dataset):
        if level in self.levels:
            self.levels[level].setParent(None)
            del self.levels[level]

    def on_export(self):
        levels = []
        for level in self.levels.values():
            if level.isChecked():
                levels.append(level.value())
        ftype = ['hdf5', 'mrc', 'tiff'][self.format.currentIndex()]
        otype = ['rdata', 'raw', 'mask', 'data'][self.combo.currentIndex()]
        owrite = self.overwrite.isChecked()
        dest = self.output.value()

        if otype == 'rdata':
            return self.save_raw_data('data', dest, ftype, owrite)

        self.launcher.setup('Exporting Levels')
        if not levels:
            self.launcher.error.emit("Please select a level!")
            return
        for level in levels:
            log.info('+ Exporting [Level {}]'.format(level))
            if otype == 'raw':
                self.save_raw(level, dest, ftype, owrite)
            elif otype == 'mask':
                self.save_mask(level, dest, ftype, owrite)
            else:
                self.save_data(level, dest, ftype, owrite, self.source.value(),
                               self.scale.isChecked(), self.invert.isChecked())
        self.launcher.cleanup()

    def save_raw(self, level, dest, ftype, owrite):
        log.info('+ Loading data into memory')
        dataset = self.LBLM.dataset(level)
        data = self.DM.load_ds(dataset)
        fname = os.path.basename(dataset)
        if isinstance(dest, str):
            outpath = os.path.join(dest, fname)
            if ftype == 'hdf5':
                attrs = self.DM.attrs(dataset)
                self.save_hdf5(outpath + '.h5', data, owrite=owrite, attrs=attrs)
            elif ftype == 'mrc':
                data = (data + 1).astype(np.int16)
                self.save_mrc(outpath + '.mrc', data, owrite=owrite)
            else:
                data = (data + 1).astype(np.int16)
                self.save_tiff(outpath + '.tif', data, owrite=owrite)
        else:
            self.launcher.error.emit("No output folder specified.")


    def save_mask(self, level, dest, ftype, owrite):
        log.info('+ Loading data into memory')
        dataset = self.LBLM.dataset(level)
        data = self.DM.load_ds(dataset)
        fname = os.path.basename(dataset)
        if isinstance(dest, str):
            for label in self.LBLM.labels(level):
                log.info('+ Saving [{}] from [Level {}]'.format(label.name, level))
                outpath = os.path.join(dest, '{}-mask{}'.format(fname, label.idx))
                mask = img_as_ubyte(data == label.idx)
                if ftype == 'hdf5':
                    self.save_hdf5(outpath + '.h5', mask, owrite=owrite)
                elif ftype == 'mrc':
                    self.save_mrc(outpath + '.mrc', mask, owrite=owrite)
                else:
                    self.save_tiff(outpath + '.tif', mask, owrite=owrite)
        else:
            self.launcher.error.emit("No output folder specified.")

    def save_data(self, level, dest, ftype, owrite, source_ds, scale=True, invert=False):
        log.info('+ Loading data into memory')
        dataset = self.LBLM.dataset(level)
        data = self.DM.load_ds(dataset)
        fname = os.path.basename(dataset)
        if isinstance(dest, str):
            source = self.DM.load_ds(source_ds)
            source[np.isinf(source)] = np.nan
            if invert:
                source = -source

            log.info('+ Extracting data stats')
            mask = ~np.isnan(source)
            amin = np.min(source[mask]); amax = np.max(source[mask]);
            if scale:
                source[mask] -= amin
                source[mask] /= (amax - amin)
                source[~mask] = 0
                amin = 0; amax = 1;
            amean = source.mean()

            fillval = amax if invert else amin

            for label in self.LBLM.labels(level):
                log.info('+ Saving [{}] from [Level {}]'.format(label.name, level))
                outpath = os.path.join(dest, '{}-data{}'.format(fname, label.idx))
                mask = (data == label.idx)
                final = np.full(source.shape, fillval, np.float32)
                final[mask] = source[mask]

                if ftype == 'hdf5':
                    attrs = self.DM.attrs(source_ds)
                    self.save_hdf5(outpath + '.h5', final, owrite=owrite, attrs=attrs)
                elif ftype == 'mrc':
                    self.save_mrc(outpath + '.mrc', final, owrite=owrite)
                else:
                    self.save_tiff(outpath + '.tif', final, owrite=owrite)
        else:
            self.launcher.error.emit("No output folder specified.")

    def save_raw_data(self, dataset, dest, ftype, owrite):
        log.info('+ Loading data into memory')
        data = self.DM.load_ds(dataset)
        data -= data.min()
        data /= data.max()

        fname = os.path.basename(dataset)

        if isinstance(dest, str):
            outpath = os.path.join(dest, fname)
            if ftype == 'hdf5':
                attrs = self.DM.attrs(dataset)
                self.save_hdf5(outpath + '.h5', data, owrite=owrite, attrs=attrs)
            elif ftype == 'mrc':
                stats = (data.min(), data.max(), data.mean())
                self.save_mrc(outpath + '.mrc', data, owrite=owrite, stats=stats)
            else:
                self.save_tiff(outpath + '.tif', data, owrite=owrite)
        else:
            self.launcher.error.emit("No output folder specified.")

    def check_owrite(self, path, flag=False):
        if flag:
            return True
        if os.path.isfile(path):
            err_msg = 'Destination dataset [{}] already exists, '\
                      'do you want to overwrite it?'.format(path)
            ans = QtWidgets.QMessageBox.question(self, "Error", err_msg,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)
            if ans == QtWidgets.QMessageBox.No:
                log.info('  -- Skipping')
                return False
        return True

    def save_hdf5(self, outpath, data, owrite=True, attrs=None):
        if not self.check_owrite(outpath, owrite):
            return
        log.info(' * Writing file [{}]'.format(outpath))
        with h5.File(outpath, 'w') as f:
            d = f.create_dataset('data', data=data)
            if attrs is not None:
                for k, v in attrs.items():
                    d.attrs[k] = v

    def save_mrc(self, outpath, data, owrite=True, stats=None):
        if not self.check_owrite(outpath, owrite):
            return

        log.info(' * Writing file [{}]'.format(outpath))
        MRC(data, stats=stats).save(outpath)

    def save_tiff(self, outpath, data, owrite=True):
        if not tiff_enabled:
            raise Exception('`tifffile` library not available')

        log.info(' * Writing file [{}]'.format(outpath))
        tiffsave(outpath, data)





import numpy as np
import logging as log

from scipy import ndimage as ndi
from itertools import combinations_with_replacement

from ..core import DataModel
from ..lib.convolutions import gconvssh, make_gaussian_1d
from ..lib._preprocess import tvdenoising
from ..lib._channels import symmetric_eigvals3S_gpu


DM = DataModel.instance()

### RAW CLAMPED DATA

def compute_threshold(data=None, params=None):
    return data

def compute_inv_threshold(data=None, params=None):
    return -data

### DENOISING

def compute_gaussian(data=None, params=None):
    sz, sy, sx = params['Sigma']
    kz = make_gaussian_1d(sz, order=0, trunc=3)
    ky = make_gaussian_1d(sy, order=0, trunc=3)
    kx = make_gaussian_1d(sx, order=0, trunc=3)

    log.info('+ Padding data')
    d, h, w = kz.size//2, ky.size//2, kx.size//2
    data = np.pad(data, ((d,d),(h,h),(w,w)), mode='reflect')

    log.info('   - Computing gaussian (radius={})'.format((d, h, w)))
    gauss = gconvssh(data, kz, ky, kx, gpu=DM.selected_gpu)

    return gauss

def compute_tvdenoise(data=None, params=None):
    return tvdenoising(data, params['Lambda'], spacing=tuple(params['Spacing']),
                       max_iter=params['Max Iter'], gpu=DM.selected_gpu)

### LOCAL STATISTICS

def compute_local_mean(data=None, params=None):
    log.info('+ Padding data')
    d, h, w = params['Radius']
    data = np.pad(data, ((d,d),(h,h),(w,w)), mode='reflect')

    log.info('+ Computing local mean features (radius={})'.format((d,h,w)))
    kernelz = np.ones(d * 2 + 1, np.float32)
    kernelz /= kernelz.sum()
    kernely = np.ones(h * 2 + 1, np.float32)
    kernely /= kernely.sum()
    kernelx = np.ones(w * 2 + 1, np.float32)
    kernelx /= kernelx.sum()

    features = gconvssh(data, kernelz, kernely, kernelx, gpu=DM.selected_gpu)

    return features

def compute_local_std(data=None, params=None):
    log.info('+ Padding data')
    d, h, w = params['Radius']
    data = np.pad(data, ((d,d),(h,h),(w,w)), mode='reflect')

    log.info('+ Computing local mean features (radius={})'.format((d,h,w)))
    kernelz = np.ones(d * 2 + 1, np.float32)
    kernelz /= kernelz.sum()
    kernely = np.ones(h * 2 + 1, np.float32)
    kernely /= kernely.sum()
    kernelx = np.ones(w * 2 + 1, np.float32)
    kernelx /= kernelx.sum()

    log.info('   - Computing local mean squared')
    meansq = gconvssh(data, kernelz, kernely, kernelx, gpu=DM.selected_gpu)
    meansq **= 2

    log.info('   - Computing local squared mean')
    data **= 2
    sqmean = gconvssh(data, kernelz, kernely, kernelx, gpu=DM.selected_gpu)
    sqmean -= meansq
    np.maximum(sqmean, 0, out=sqmean)
    np.sqrt(sqmean, sqmean) # inplace

    return sqmean

def compute_local_centering(data=None, params=None):
    mean = compute_local_mean(data=data, params=params)
    return data - mean

def compute_local_norm(data=None, params=None):
    mean = compute_local_mean(data=data, params=params)
    std = compute_local_std(data=data, params=params)
    mask = ~np.isclose(std, 0)
    result = np.zeros_like(data)
    result[mask] = (data[mask] - mean[mask]) / std[mask]
    return result

def compute_local_magnitude(data=None, params=None):
    gz, gy, gx = np.gradient(data)
    mag = np.sqrt(gz**2 + gy**2 + gx**2)
    mean = compute_local_mean(data=mag, params=params)
    return mean

### GAUSSIAN

def compute_gaussian_centering(data=None, params=None):
    g1 = compute_gaussian(data=data, params=params)
    return data - g1

def compute_gaussian_normalization(data=None, params=None):
    g1 = compute_gaussian(data=data, params=params)
    num = data - g1
    g2 = compute_gaussian(data=num**2, params=params)
    den = np.sqrt(g2)
    mask = ~np.isclose(den, 0)
    result = np.zeros_like(data)
    result[mask] = num[mask] / den[mask]
    return result

def compute_gaussian_magnitude(data=None, params=None):
    sz, sy, sx = params['Sigma']
    out = []

    for oz, oy, ox in [(1,0,0),(0,1,0),(0,0,1)]:
        szz = sz if oz == 0 else sz * 3
        syy = sy if oy == 0 else sy * 3
        sxx = sx if ox == 0 else sx * 3
        kz = make_gaussian_1d(szz, order=oz, trunc=3)
        ky = make_gaussian_1d(syy, order=oy, trunc=3)
        kx = make_gaussian_1d(sxx, order=ox, trunc=3)

        log.info('+ Padding data')
        d, h, w = kz.size//2, ky.size//2, kx.size//2
        tmp = np.pad(data, ((d,d),(h,h),(w,w)), mode='reflect')

        log.info('   - Computing convolutions (radius={})'.format((d, h, w)))
        gauss = gconvssh(tmp, kz, ky, kx, gpu=DM.selected_gpu)
        out.append(gauss)

    out = np.sqrt(out[0]**2 + out[1]**2 + out[2]**2)

    return out

### BLOB DETECTION

def compute_difference_gaussians(data=None, params=None):
    sigma1 = params['Sigma']
    sigma2 = tuple(np.asarray(sigma1) * params['Sigma Ratio'])
    if 'Response' in params and params['Response'] == 'Dark':
        data *= -1
    g1 = compute_gaussian(data=data, params=dict(Sigma=sigma1))
    g2 = compute_gaussian(data=data, params=dict(Sigma=sigma2))
    response = g1 - g2
    if 'Threshold' in params and params['Threshold']:
        response[response < 0] = 0
    return response


def compute_laplacian_gaussian(data=None, params=None):
    sz, sy, sx = params['Sigma']
    out = np.zeros_like(data)

    for i, (oz, oy, ox) in enumerate([(2,0,0),(0,2,0),(0,0,2)]):
        kz = make_gaussian_1d(sz, order=oz, trunc=3)
        ky = make_gaussian_1d(sy, order=oy, trunc=3)
        kx = make_gaussian_1d(sx, order=ox, trunc=3)

        if 'Response' in params and params['Response'] == 'Bright':
            if i == 0: kz *= -1
            if i == 1: ky *= -1
            if i == 2: kx *= -1

        log.info('+ Padding data')
        d, h, w = kz.size//2, ky.size//2, kx.size//2
        tmp = np.pad(data, ((d,d),(h,h),(w,w)), mode='reflect')

        log.info('   - Computing convolutions (radius={})'.format((d, h, w)))
        gauss = gconvssh(tmp, kz, ky, kx, gpu=DM.selected_gpu)
        out += gauss

    if 'Threshold' in params and params['Threshold']:
        out[out < 0] = 0

    return out


### HESSIAN

def compute_hessian(data=None, params=None):
    log.info('+ Computing Hessian Matrix')
    gaussian_filtered = compute_gaussian(data=data, params=params)
    gradients = [np.gradient(gaussian_filtered, axis=i) for i in range(3)]
    axes = range(data.ndim)
    H_elems = [np.gradient(gradients[2-ax0], axis=2-ax1)
               for ax0, ax1 in combinations_with_replacement(axes, 2)]
    sigma = max(params['Sigma'])
    if sigma > 1:
        H_elems = [elem * sigma**2 for elem in H_elems]
    return H_elems

def compute_hessian_determinant(data=None, params=None):
    Hxx, Hxy, Hxz, Hyy, Hyz, Hzz = compute_hessian(data=data, params=params)

    log.info('+ Computing Hessian Determinant')
    det =  Hxx * (Hyy*Hzz - Hyz*Hyz) \
         - Hxy * (Hxy*Hzz - Hyz*Hxz) \
         + Hxz * (Hxy*Hyz - Hyy*Hxz)
    if 'Response' in params and params['Response'] == 'Bright':
        return -det
    return det

def hessian_eigvals(data=None, params=None, correct=False, doabs=False): # TODO: GPU THIS
    H = compute_hessian(data=data, params=params)
    if correct:
        s = max(params['Sigma'])**2
        Hxx, Hxy, Hxz, Hyy, Hyz, Hzz = [h * s for h in H]
    else:
        Hxx, Hxy, Hxz, Hyy, Hyz, Hzz = H
    log.info('+ Computing Hessian Eigenvalues')
    R = symmetric_eigvals3S_gpu(Hxx, Hxy, Hxz, Hyy, Hyz, Hzz, doabs=doabs, gpu=DM.selected_gpu)
    return R

def compute_hessian_eigvals(data=None, params=None, correct=False):
    R = hessian_eigvals(data=data, params=params, correct=correct)
    return R[..., params['Eigen Value']].copy()

### STRUCTURE TENSOR

def compute_structure_tensor(data=None, params=None):
    log.info('+ Computing Structure Tensor')
    p1 = dict(Sigma=params['Sigma Deriv'])
    p2 = dict(Sigma=params['Sigma Area'])

    gradients = np.gradient(data)
    H_elems = [compute_gaussian(gradients[2-ax0] * gradients[2-ax1], params=p2)
               for ax0, ax1 in combinations_with_replacement(range(3), 2)]
    return H_elems


def compute_structure_tensor_determinant(data=None, params=None):
    Sxx, Sxy, Sxz, Syy, Syz, Szz = compute_structure_tensor(data=data, params=params)

    log.info('+ Computing Structure Tensor Determinant')
    return  Sxx * (Syy*Szz - Syz*Syz) \
          - Sxy * (Sxy*Szz - Syz*Sxz) \
          + Sxz * (Sxy*Syz - Syy*Sxz)


def compute_structure_tensor_eigvals(data=None, params=None):
    Sxx, Sxy, Sxz, Syy, Syz, Szz = compute_structure_tensor(data=data, params=params)
    log.info('+ Computing Structure Tensor Eigenvalues')
    R = symmetric_eigvals3S_gpu(Sxx, Sxy, Sxz, Syy, Syz, Szz, gpu=DM.selected_gpu)
    return R[..., params['Eigen Value']].copy()


def compute_gaussian_scale_invariant(data=None, params=None):
    sigma = np.asarray(params['Init Sigma'], dtype=np.float32)
    sigincr = params['Sigma Incr']
    max_sigma = params['Max Sigma']
    response = params['Response']

    result = np.full(data.shape, -np.inf, data.dtype)
    num_scales = 0

    while max(sigma) < max_sigma:
        tmp = compute_gaussian(data=data, params=dict(Sigma=sigma))
        if response == 'Min':
            result = np.minimum(result, tmp)
        elif response == 'Max':
            result = np.maximum(result, tmp)
        else:
            result += tmp
            num_scales += 1
        sigma += sigincr

    if response == 'Avg' and num_scales > 0:
        result /= num_scales

    return result

### SCALE INVARIANT DoG

def compute_sidog(data=None, params=None):
    sigma = np.asarray(params['Init Sigma'], dtype=np.float32)
    max_sigma = params['Max Sigma']
    sratio = params['Sigma Ratio']
    response = params['Response']

    result = np.full(data.shape, -np.inf, data.dtype)

    if 'Response' in params and params['Response'] == 'Dark':
        data *= -1

    gauss = compute_gaussian(data, params=dict(Sigma=sigma))

    while max(sigma) * sratio < max_sigma:
        prev_sigma = sigma.copy()

        sigma *= sratio
        gauss1 = compute_gaussian(data, params=dict(Sigma=sigma))

        tmp = (gauss - gauss1) * max(prev_sigma)
        np.maximum(result, tmp, out=result)
        gauss = gauss1

    if 'Threshold' in params and params['Threshold']:
        result[result < 0] = 0

    return result

### SCALE INVARIANT LAPLACIAN

def compute_silaplacian(data=None, params=None):
    sigma = np.asarray(params['Init Sigma'], dtype=np.float32)
    max_sigma = params['Max Sigma']
    sincr = params['Sigma Incr']

    result = np.full(data.shape, -np.inf, data.dtype)

    while max(sigma) < max_sigma:
        params = dict(Sigma=sigma, Response=params['Response'])
        tmp = compute_laplacian_gaussian(data=data, params=params) * max(sigma)**2
        np.maximum(result, tmp, out=result)
        sigma += sincr

    if 'Threshold' in params and params['Threshold']:
        result[result < 0] = 0

    return result

### SCALE INVARIANT HESSIAN DETERMINANT

def compute_sihessian_det(data=None, params=None):
    sigma = np.asarray(params['Init Sigma'], dtype=np.float32)
    max_sigma = params['Max Sigma']
    sincr = params['Sigma Incr']

    result = np.full(data.shape, -np.inf, data.dtype)

    while max(sigma) < max_sigma:
        params = dict(Sigma=sigma, Response=params['Response'])
        tmp = compute_hessian_determinant(data=data, params=params)
        np.maximum(result, tmp, out=result)
        sigma += sincr

    return result

### FRANGI FILTER

def compute_frangi(data=None, params=None):
    sigma = np.asarray(params['Init Sigma'], dtype=np.float32)
    max_sigma = params['Max Sigma']
    sincr = params['Sigma Incr']

    result = None

    while max(sigma) < max_sigma:
        R = hessian_eigvals(data=data, params=dict(Sigma=sigma),
                            correct=True, doabs=True)
        e1 = R[..., 0]
        e2 = R[..., 1]
        e3 = R[..., 2]

        ae1 = np.abs(e1)
        ae2 = np.abs(e2)
        ae3 = np.abs(e3)

        ae1sq = ae1 * ae1
        ae2sq = ae2 * ae2
        ae3sq = ae3 * ae3

        Ra = (ae2sq / ae3sq)
        Rb = (ae1sq / (ae2 * ae3))
        S = ae1sq + ae2sq + ae3sq

        A = B = 2 * (params['Lamda']**2)
        C = 2 * S.max()

        expRa = 1 - np.exp(-Ra / A)
        expRb =     np.exp(-Rb / B)
        expS  = 1 - np.exp(-S  / C)

        tmp = expRa * expRb * expS

        if params['Response'] == 'Dark':
            tmp[e2 < 0] = 0; tmp[e3 < 0] = 0;
        else:
            tmp[e2 > 0] = 0; tmp[e3 > 0] = 0;

        tmp[np.isnan(tmp)] = 0

        if result is None:
            result = tmp
        else:
            np.maximum(result, tmp, out=result)

        sigma += sincr

    return result


def compute_relu(data=None, params=None):
    rtype = params['Type']
    alpha = params['Alpha']

    if rtype == 'Standard':
        np.maximum(0, data, out=data)
    elif rtype == 'Noisy':
        gnoise = np.random.randn(*data.shape).astype(np.float32)
        np.maximum(0, data + alpha * gnoise, out=data)
    else:
        data[data < 0] *= alpha

    return data

###############################################################################
# MAIN FUNC
###############################################################################

fnmap = {
    'thresh'            : compute_threshold,
    'inv_thresh'        : compute_inv_threshold,

    'gauss'             : compute_gaussian,
    'tv'                : compute_tvdenoise,

    'local_mean'        : compute_local_mean,
    'local_std'         : compute_local_std,
    'local_center'      : compute_local_centering,
    'local_norm'        : compute_local_norm,
    'local_mag'         : compute_local_magnitude,

    'gauss_center'      : compute_gaussian_centering,
    'gauss_norm'        : compute_gaussian_normalization,
    'gauss_mag'         : compute_gaussian_magnitude,

    'dog'               : compute_difference_gaussians,
    'log'               : compute_laplacian_gaussian,

    'hessian_det'       : compute_hessian_determinant,
    'hessian_eig'       : compute_hessian_eigvals,

    'structure_det'     : compute_structure_tensor_determinant,
    'structure_eig'     : compute_structure_tensor_eigvals,

    'gauss_scale3d'     : compute_gaussian_scale_invariant,
    'si_dog'            : compute_sidog,
    'si_laplacian'      : compute_silaplacian,
    'si_hessian_det'    : compute_sihessian_det,
    'frangi'            : compute_frangi,

    'relu'              : compute_relu
}

def compute_channel(source=None, clamp=None, feature=None,
                    idx=None, name=None, params=None, out=None):
    result = None

    log.info('+ Loading data into memory')
    data = DM.load_slices(source)

    if clamp is not None:
        log.info('+ Clamping data')
        t1, t2 = clamp
        data[data < t1] = t1
        data[data > t2] = t2
        data -= t1
        data /= (t2 - t1)
    elif feature != 'relu':
        log.info('+ Rescaling data')
        data -= data.min()
        data /= data.max()

    if feature in fnmap:
        log.info('+ Computing feature {}'.format(feature))
        result = fnmap[feature](data=data, params=params)

        log.info('+ Normalizing feature')
        result -= result.mean()
        std = result.std()
        if std >= 1e-5:
            result /= std

        log.info('+ Calculating feature statistics')
        amin, amax = result.min(), result.max()
        evmin, evmax = np.percentile(result, (1., 99.))

        log.info('+ Saving feature into disk {}'.format(result.shape))
        params['source'] = source
        params['feature_idx'] = idx
        params['feature_name'] = name
        params['feature_type'] = feature
        params['vmin'] = amin
        params['vmax'] = amax
        params['evmin'] = evmin
        params['evmax'] = evmax
        params['default_evmin'] = evmin
        params['default_evmax'] = evmax
        params['active'] = True
        DM.write_slices(out, result, params=params)
        return out, idx, params

    return None


def compute_all_channel(features=None):
    output = []
    for feature in features:
        log.info('\n## Computing Feature {}'.format(feature['feature']))
        output.append(compute_channel(**feature))
    return output




import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

import logging as log

from matplotlib.colors import ListedColormap

from ..core import Launcher
from ..widgets import HSize3D, HWidgets, PLineEdit, SourceCombo, ActionButton
from .base import Plugin
from ..core import DataModel, LayerManager
from .. import actions as ac

from ..lib._features import find_boundaries

class MegaVoxels(QtWidgets.QWidget):

    name = 'MegaVoxels'

    def __init__(self, parent=None):
        super(MegaVoxels, self).__init__(parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.setLayout(QtWidgets.QVBoxLayout())

        self.source_combo = SourceCombo()
        self.layout().addWidget(HWidgets(None, 'Source:', self.source_combo, stretch=[1]))

        self.txt_lambda = PLineEdit(0.1, parse=float)
        self.txt_bins = PLineEdit(20, parse=int)
        dummy = HWidgets(QtWidgets.QLabel('Lamda:'), self.txt_lambda,
                         QtWidgets.QLabel('NumBins:'), self.txt_bins,
                         stretch=[1,0,1,0])
        self.layout().addWidget(dummy)

        self.txt_gamma = PLineEdit(None, parse=self.parse_gamma)
        dummy = HWidgets(None, QtWidgets.QLabel('Gamma:'), self.txt_gamma,
                         stretch=[1,0,0])
        self.layout().addWidget(dummy)

        self.btn_apply = ActionButton('Apply')

        dummy = HWidgets(QtWidgets.QWidget(), self.btn_apply,\
                         stretch=[1,0,])

        self.layout().addWidget(dummy)

        self.btn_apply.clicked.connect(self.apply_megavoxels)

    def parse_gamma(self, val):
        try:
            val = int(val)
            return val
        except:
            return None

    def apply_megavoxels(self):
        target = self.source_combo.value()
        ds = 'data/{}'.format(target)

        if self.DM.svtotal is None or not self.DM.has_ds('supervoxels/supervoxels'):
            Launcher.instance().error.emit('SuperVoxels need to be computed first')
            return

        self.params = {
            'lamda' : self.txt_lambda.value(),
            'nbins' : self.txt_bins.value(),
            'gamma' : self.txt_gamma.value()
        }

        in_data = target
        in_sv = 'supervoxels/supervoxels'

        out_mv = 'megavoxels/megavoxels'
        out_mvindex = 'megavoxels/megavoxels_idx'
        out_mvtable = 'megavoxels/megavoxels_table'
        self.DM.remove_dataset(out_mv)
        self.DM.remove_dataset(out_mvindex)
        self.DM.remove_dataset(out_mvtable)

        Launcher.instance().run(ac.create_megavoxels,
                                dataset=in_data,
                                splabels=in_sv,
                                num_sps=self.DM.svtotal,
                                out_mv=out_mv,
                                out_mvindex=out_mvindex,
                                out_mvtable=out_mvtable,
                                cb=self.on_megavoxels,
                                caption='Creating MegaVoxels..',
                                **self.params)

    def on_megavoxels(self, params):
        megavoxels, total_mv, sortindex, sorttable = params
        self.update_megavoxel_layer(megavoxels, sortindex, sorttable, total_mv, visible=True)

    def update_megavoxel_layer(self, mvlabels, sortindex, sorttable, mvtotal, visible=False):
        self.DM.mvlabels = mvlabels
        self.DM.mvindex = sortindex
        self.DM.mvtable = sorttable
        self.DM.mvtotal = mvtotal

        if self.LM.isin('MegaVoxels', 'Super-Regions'):
            self.LM.get('MegaVoxels', 'Super-Regions').data = self.DM.mvlabels
            self.LM.setVisible('MegaVoxels', 'Super-Regions', True)
        else:
            cmap = ListedColormap(['Maroon'])
            self.LM.addLayer(self.DM.mvlabels, 'MegaVoxels', level='Super-Regions',
                             pre=find_boundaries, cmap=cmap, background=0,
                             vmin=0.1, vmax=1, visible=visible)
        self.LM.update()

    def update_params(self, source, total_mv, lamda, nbins, gamma):
        log.info('  * {} megavoxels loaded'.format(total_mv))
        self.source_combo.loadSource(source)
        self.txt_lambda.setText(str(lamda))
        self.txt_bins.setText(str(nbins))
        self.txt_gamma.setText(str(gamma))





import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

from six.moves.queue import Queue

from .base import Plugin
from ..widgets import HWidgets, LCheckBox, RCheckBox, PLineEdit, \
                      HeaderLabel, SubHeaderLabel, TComboBox, \
                      RoundedWidget, HSize3D, SourceCombo, BLabel, \
                      ActionButton, SectionCombo
from ..core import DataModel, LayerManager, Launcher
from .. import actions as ac

import logging as log

class Int3d: pass
class Float3d: pass

class Choice(object):
    def __init__(self, options):
        self.options = options

available_features = [
    ('Raw', None, None),
    ('Threshold', 'thresh', []),
    ('Invert Threshold', 'inv_thresh', []),
    ##################################################
    ('Denoising', None, None),
    ('Gaussian Filter', 'gauss', [
        ('Sigma', Float3d, 2.),
    ]),
    ('Total Variation', 'tv', [
        ('Spacing', Float3d, 1.),
        ('Lambda', float, 10.),
        ('Max Iter', int, 100)
    ]),
    ##################################################
    ('Local Features', None, None),
    ('Local Mean', 'local_mean', [
        ('Radius', Int3d, 4),
    ]),
    ('Local Standard Deviation', 'local_std', [
        ('Radius', Int3d, 4),
    ]),
    ('Local Centering', 'local_center', [
        ('Radius', Int3d, 4),
    ]),
    ('Local Normalization', 'local_norm', [
        ('Radius', Int3d, 4),
    ]),
    ('Local Gradient Magnitude', 'local_mag', [
        ('Radius', Int3d, 4),
    ]),
    ##################################################
    ('Gaussian Features', None, None),
    ('Gaussian Centering', 'gauss_center', [
        ('Sigma', Float3d, 2.),
    ]),
    ('Gaussian Normalization', 'gauss_norm', [
        ('Sigma', Float3d, 2.),
    ]),
    ('Gaussian Gradient Magnitude', 'gauss_mag', [
        ('Sigma', Float3d, 2.),
    ]),
    ##################################################
    ('Blob Detection', None, None),
    ('Difference of Gaussian', 'dog', [
        ('Sigma', Float3d, 2.),
        ('Sigma Ratio', float, 1.6),
        ('Threshold', bool, False),
        ('Response', Choice(['Bright', 'Dark']), 'Bright'),
    ]),
    ('Laplacian of Gaussian', 'log', [
        ('Sigma', Float3d, 2.),
        ('Threshold', bool, False),
        ('Response', Choice(['Bright', 'Dark']), 'Bright')
    ]),
    ('Determinant of Hessian', 'hessian_det', [
        ('Sigma', Float3d, 2.),
    ]),
    ('Determinant of Struct. Tensor', 'structure_det', [
        ('Sigma Deriv', Float3d, 1.),
        ('Sigma Area', Float3d, 3.),
    ]),
    ##################################################cd
    ('Texture and Structure', None, None),
    ('Hessian Eigenvalues', 'hessian_eig', [
        ('Sigma', Float3d, 1.),
        ('Eigen Value', Choice([0,1,2]), 0),
    ]),
    ('Struct. Tensor Eigenvalues', 'structure_eig', [
        ('Sigma Deriv', Float3d, 1.),
        ('Sigma Area', Float3d, 1.),
        ('Eigen Value', Choice([0,1,2]), 0),
    ]),
    ##################################################
    ('Robust Features', None, None),
    ('(SI) Gaussian', 'gauss_scale3d', [
        ('Init Sigma', Float3d, 1),
        ('Sigma Incr', float, 2.),
        ('Max Sigma', float, 10.),
        ('Response', Choice(['Min', 'Max', 'Avg']), 'Max')
    ]),
    #('Derivative Rotation Invariant', 'gauss_ori3d', [
    #	('Sigma', Float3d, 3.),
    #	('Num Orientations', Choice([2,4,8,16]), 8),
    #	('Response', Choice(['Min', 'Max', 'Avg']), 'Max')
    #]),
    ('(SI) Difference of Gaussians', 'si_dog', [
        ('Init Sigma', Float3d, 1),
        ('Sigma Ratio', float, 1.6),
        ('Max Sigma', float, 15.),
        ('Threshold', bool, False),
        ('Response', Choice(['Bright', 'Dark']), 'Bright')
    ]),
    ('(SI) Laplacian of Gaussian', 'si_laplacian', [
        ('Init Sigma', Float3d, 1),
        ('Sigma Incr', float, 2.),
        ('Max Sigma', float, 10.),
        ('Threshold', bool, False),
        ('Response', Choice(['Bright', 'Dark']), 'Bright')
    ]),
    ('(SI) Determinant of Hessian', 'si_hessian_det', [
        ('Init Sigma', Float3d, 1),
        ('Sigma Incr', float, 2.),
        ('Max Sigma', float, 10.),
        ('Response', Choice(['Bright', 'Dark']), 'Bright')
    ]),
    ('(SI) Frangi Filter', 'frangi', [
        ('Init Sigma', Float3d, 1),
        ('Sigma Incr', float, 2.),
        ('Max Sigma', float, 10.),
        ('Lamda', float, 0.5),
        ('Response', Choice(['Bright', 'Dark']), 'Bright')
    ]),
    ##################################################
    ('Activation Layers', None, None),
    ('Rectified Linear Unit', 'relu', [
        ('Type', Choice(['Standard', 'Noisy', 'Leaky']), 'Standard'),
        ('Alpha', float, 0.01)
    ]),
]

available_descriptors = [
    'Mean',
    'Feature Histogram',
    'Texton Histogram',
    'Covariance',
    'SigmaSet'
]

class FeatureRow(RoundedWidget):

    toggled = QtCore.pyqtSignal(int, bool)
    delete = QtCore.pyqtSignal(int)
    compute = QtCore.pyqtSignal(int, str, str, str, dict)

    def __init__(self, idx, ftype=1, fname=None, active=None,
                 load_params={}, parent=None):
        super(FeatureRow, self).__init__(parent=parent, color=None,
                                         bg='#cde5e5', width=0)

        self.DM = DataModel.instance()

        self.idx = idx
        self.feature = available_features[ftype]
        if fname is None:
            if self.feature[1] is None:
                raise ValueError("No feature selected!")
            else:
                self.fname = '{}_{}'.format(idx, self.feature[1])
        else:
            self.fname = fname
        self.dsname = 'channels/{}'.format(self.fname)
        params = dict(active=False, feature_idx=self.idx,
                      feature_type=self.feature[1], feature_name=self.fname)
        if len(params) > 0:
            self.DM.create_empty_dataset(self.dsname, shape=self.DM.data_shape,
                                         dtype=np.float32, params=params,
                                         fillvalue=np.nan)

        self.chk_compute = QtWidgets.QCheckBox()
        self.btn_delete = QtWidgets.QPushButton('X')
        self.btn_delete.setMaximumWidth(25)
        self.btn_delete.setMaximumHeight(30)
        self.btn_delete.setMinimumHeight(30)
        self.btn_delete.setMinimumWidth(25)

        self.txt_name = QtWidgets.QLabel('[{}] {}'.format(self.idx, self.feature[0]))
        self.txt_name.setStyleSheet('background-color: #6DC7C7; color: #006161;'
                                    'padding-left: 10px;')

        self.btn_compute = QtWidgets.QPushButton(u'\u2713')
        self.btn_compute.setContentsMargins(0,0,0,0)
        self.btn_compute.setMaximumWidth(25)
        self.btn_compute.setCheckable(True)
        if active == True:
            self.btn_compute.setChecked(True)

        self.btn_show = QtWidgets.QPushButton(u'\u25B2')
        self.btn_show.setMaximumWidth(25)
        self.btn_show.setMaximumHeight(30)
        self.btn_show.setMinimumHeight(30)
        self.btn_show.setMinimumWidth(25)
        self.visible = True

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        vbox.addWidget(HWidgets(self.chk_compute, self.btn_delete,
                                self.txt_name,
                                self.btn_compute, self.btn_show,
                                stretch=[0,0,1,0,0]))

        self.txt_name.mousePressEvent = self.toggleParams
        self.btn_show.clicked.connect(self.toggleParams)
        self.btn_delete.clicked.connect(self.on_delete)
        self.btn_compute.clicked.connect(self.on_compute)

        self.param_container = QtWidgets.QWidget()
        vbox.addWidget(self.param_container)

        vbox = QtWidgets.QVBoxLayout(self.param_container)
        vbox.setContentsMargins(0,0,0,0)

        self.source_combo = SourceCombo(ignore=self.fname) # ignore self
        self.clamp_chk = LCheckBox('Clamp')
        if 'clamp' in load_params:
            self.clamp_chk.setChecked(load_params['clamp'])
        else:
            self.clamp_chk.setChecked(False)
        if available_features[ftype][1] == 'thresh':
            self.clamp_chk.setChecked(True)
            self.clamp_chk.setEnabled(False)
        if 'source' in load_params:
            self.source_combo.loadSource(load_params['source'])
        vbox.addWidget(HWidgets(self.clamp_chk, None, self.source_combo,
                                stretch=[0,1,0,0]))

        self.isChecked = self.chk_compute.isChecked
        self.setChecked = self.chk_compute.setChecked

        self.params = []
        self.init_params(vbox, load_params)
        self.showParams(False)

    def init_params(self, vbox, load_params):
        for param in self.feature[2]:
            if param[1] == bool:
                curr = QtWidgets.QCheckBox()
                curr.setChecked(param[2])
                curr.value = curr.isChecked
                if param[0] in load_params:
                    curr.setChecked(load_params[param[0]])
            elif param[1] == Float3d:
                curr = HSize3D('', default=(param[2],param[2],param[2]),
                               parse=float, txtwidth=10, coordwidth=50)
                if param[0] in load_params:
                    curr.setValue(load_params[param[0]])
            elif param[1] == Int3d:
                curr = HSize3D('', default=(param[2],param[2],param[2]),
                               parse=int, txtwidth=10, coordwidth=50)
                if param[0] in load_params:
                    curr.setValue(load_params[param[0]])
            elif type(param[1]) == Choice:
                curr = TComboBox('', param[1].options, parse=type(param[1].options[0]))
                curr.setCurrentIndex(param[1].options.index(param[2]))
                if param[0] in load_params:
                    curr.setCurrentIndex(param[1].options.index(load_params[param[0]]))
            else:
                curr = PLineEdit(parse=param[1], default=param[2])
                if param[0] in load_params:
                    curr.setText(str(load_params[param[0]]))

            vbox.addWidget(HWidgets(None, param[0], curr, stretch=[1,0,0]))
            self.params.append((param[0], curr))

    def showParams(self, visible):
        self.visible = visible
        if self.visible:
            self.btn_show.setText(u'\u25BC')
        else:
            self.btn_show.setText(u'\u25B2')
        self.param_container.setVisible(self.visible)

    def toggleParams(self, dummy):
        self.showParams(not self.visible)
        self.toggled.emit(self.idx, self.visible)

    def on_delete(self):
        self.delete.emit(self.idx)

    def get_params(self):
        params = {k: v.value() for k, v in self.params}
        params['source'] = self.source_combo.value()
        params['clamp'] = self.clamp_chk.isChecked()
        return params

    def on_compute(self):
        params = self.get_params()
        self.compute.emit(self.idx, self.txt_name.text(), self.fname,
                          self.feature[1], params)

    def setComputed(self, bol):
        self.btn_compute.setChecked(bol)


class FeatureChannels(Plugin):

    name = 'Feature Channels'

    def __init__(self, parent=None):
        super(FeatureChannels, self).__init__(ptype=Plugin.Plugin, parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.launcher = Launcher.instance()

        self.addWidget(HeaderLabel('Feature Channels'))
        #self.cmb_features = QtWidgets.QComboBox(self)
        self.cmb_features = SectionCombo('Select Feature', parent=self)
        self.cmb_features.setMinimumWidth(280)
        for i, feature in enumerate(available_features):
            self.cmb_features.addItem(feature[0], section=feature[2] is None)

        self.btn_add = ActionButton('Add')

        dummy = QtWidgets.QWidget(self)
        vbox = QtWidgets.QVBoxLayout(dummy)
        dummy.setLayout(vbox)
        vbox.addWidget(HWidgets(self.cmb_features, self.btn_add, stretch=[1,0],
                                parent=self))

        self.chk_all = QtWidgets.QCheckBox("Select all features")
        self.chk_all.stateChanged.connect(self.on_check_all)
        self.cmp_all = QtWidgets.QPushButton("Compute features")
        self.cmp_all.setMinimumWidth(150)
        self.cmp_all.clicked.connect(self.on_compute_all)
        vbox.addWidget(HWidgets(self.chk_all, None, self.cmp_all, stretch=[0,1,0]))

        self.addWidget(dummy)

        spacer = QtWidgets.QWidget()
        spacer.setMinimumHeight(10)
        spacer.setMaximumHeight(10)
        vbox.addWidget(spacer)

        vbox.addWidget(SubHeaderLabel('Available Channels'))

        dummy = QtWidgets.QWidget()
        self.feature_container = QtWidgets.QVBoxLayout()
        dummy.setLayout(self.feature_container)
        self.addWidget(dummy)

        self.btn_add.clicked.connect(self.on_add_feature)
        self.DM.clf_channel_computed.connect(self.load_channel)

        self.features = {}
        self.num_features = 0

    def ftype2idx(self, ftype):
        for i, f in enumerate(available_features):
            if f[1] == ftype:
                return i
        return -1

    def load_channel(self, fidx, fname, ftype, active, params):
        ftype = self.ftype2idx(ftype)
        if ftype < 0:
            log.info('[[[Error]]] feature {} could not be loaded'.format(fname))
            return
        try:
            feature_row = FeatureRow(fidx, ftype, fname=fname, active=active,
                                 load_params=params, parent=self)
        except ValueError as e:
            log.error("Creating feature failed: {}".format(e))
            return
        feature_row.toggled.connect(self.on_toggle_feature)
        feature_row.delete.connect(self.on_delete_feature)
        feature_row.compute.connect(self.on_compute_feature)
        self.feature_container.addWidget(feature_row)

        self.features[fidx] = feature_row
        self.on_toggle_feature(-1, True)

        if self.num_features <= fidx:
            self.num_features = fidx + 1

    def on_add_feature(self):
        ftype = self.cmb_features.currentIndex()
        try:
            feature_row = FeatureRow(self.num_features, ftype, parent=self)
        except ValueError as e:
            log.error("Creating feature failed: {}".format(e))
            return
        feature_row.toggled.connect(self.on_toggle_feature)
        feature_row.delete.connect(self.on_delete_feature)
        feature_row.compute.connect(self.on_compute_feature)

        self.feature_container.addWidget(feature_row)

        self.features[self.num_features] = feature_row
        self.on_toggle_feature(-1, True)

        self.num_features += 1

    def on_toggle_feature(self, idx, visible):
        if visible:
            for k, v in self.features.items():
                if k != idx:
                    v.showParams(False)

    def on_delete_feature(self, idx):
        channel_name = self.features[idx].dsname
        self.DM.remove_dataset(channel_name)
        self.features[idx].setParent(None)
        self.DM.channel_removed.emit(channel_name)
        if idx == self.num_features - 1:
            self.num_features -= 1

    def on_check_all(self, state):
        check = (state == QtCore.Qt.Checked)
        for feature in self.features.values():
            feature.setChecked(check)

    def on_compute_feature(self, idx, feature_name, feature_id, feature_desc, params):
        source = params['source']
        clamp = None
        if params['clamp'] == True:
            attrs = self.DM.attrs(source)
            clamp = attrs['evmin'], attrs['evmax']
        out = 'channels/{}'.format(feature_id)

        self.current_name = out

        self.launcher.run(ac.compute_channel, name=feature_name,
                          source=source, clamp=clamp, out=out, idx=idx,
                          feature=feature_desc, params=params,
                          caption='Computing Feature: {}'.format(feature_desc),
                          cb=self.on_features_computed)

    def on_compute_all(self):
        features = []
        for k, feature in self.features.items():
            if not feature.isChecked():
                continue
            params = feature.get_params()
            source = params['source']
            clamp = None
            if params['clamp'] == True:
                attrs = self.DM.attrs(source)
                clamp = attrs['evmin'], attrs['evmax']
            out = feature.dsname
            idx = feature.idx
            feature_name = feature.txt_name.text()
            feature_desc = feature.feature[1]

            features.append({
                'name'          : feature_name,
                'source'        : source,
                'clamp'         : clamp,
                'params'        : params,
                'out'           : out,
                'idx'           : idx,
                'feature'       : feature_desc
            })

        self.launcher.run(ac.compute_all_channel, features=features,
                          caption='Computing Multiple Feature',
                          cb=self.on_all_computed)

    def on_all_computed(self, results):
        for result in results:
            if result is None:
                continue
            out, idx, params = result
            self.DM.channel_computed.emit(out, params)
            self.features[idx].setComputed(True)


    def on_features_computed(self, result):
        if result is None:
            QtWidgets.QMessageBox.critical(self, 'Error', 'Feature not available')
        else:
            out, idx, params = result
            self.DM.channel_computed.emit(out, params)
            self.features[idx].setComputed(True)
            self.DM.update_channel.emit(out)



import numpy as np
import os
import os.path as op
from ..qt_compat import QtGui, QtCore, QtWidgets

import logging as log

from matplotlib.colors import ListedColormap

from ..core import Launcher
from ..widgets import LCheckBox, HWidgets, BLabel, PLineEdit, TComboBox, \
    HeaderLabel, SubHeaderLabel, CSlider, ColorButton, \
    HSize3D, ActionButton
from .base import Plugin, PredWidgetBase
from ..core import DataModel, LayerManager, LabelManager
from .. import actions as ac


class EnsembleWidget(QtWidgets.QWidget):
    train_predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(EnsembleWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Ensemble Type:', [
            'Random Forest',
            'ExtraRandom Forest',
            'AdaBoost',
            'GradientBoosting'
        ])
        self.type_combo.currentIndexChanged.connect(self.on_ensemble_changed)
        vbox.addWidget(self.type_combo)

        self.ntrees = PLineEdit(100, parse=int)
        self.depth = PLineEdit(None, parse=int)
        self.lrate = PLineEdit(1., parse=float)
        self.subsample = PLineEdit(1., parse=float)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('# Trees:'),
                                self.ntrees,
                                QtWidgets.QLabel('Max Depth:'),
                                self.depth,
                                stretch=[0, 1, 0, 1]))

        vbox.addWidget(HWidgets(QtWidgets.QLabel('Learn Rate:'),
                                self.lrate,
                                QtWidgets.QLabel('Subsample:'),
                                self.subsample,
                                stretch=[0, 1, 0, 1]))

        self.btn_train_predict = ActionButton('Train && Predict')
        self.btn_train_predict.clicked.connect(self.on_train_predict_clicked)
        self.n_jobs = PLineEdit(1, parse=int)
        vbox.addWidget(HWidgets('Num Jobs', self.n_jobs, None, self.btn_train_predict,
                                stretch=[0, 0, 1, 0]))


    def on_ensemble_changed(self, idx):
        if idx == 2:
            self.ntrees.setDefault(50)
        else:
            self.ntrees.setDefault(100)

        if idx == 3:
            self.lrate.setDefault(0.1)
            self.depth.setDefault(3)
        else:
            self.lrate.setDefault(1.)
            self.depth.setDefault(None)

    def on_train_predict_clicked(self):
        ttype = ['rf', 'erf', 'ada', 'gbf']
        params = {
            'clf': 'ensemble',
            'type': ttype[self.type_combo.currentIndex()],
            'n_estimators': self.ntrees.value(),
            'max_depth': self.depth.value(),
            'learning_rate': self.lrate.value(),
            'subsample': self.subsample.value(),
            'n_jobs': self.n_jobs.value()
        }
        self.train_predict.emit(params)


class SVMWidget(QtWidgets.QWidget):
    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(SVMWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Kernel Type:', [
            'linear',
            'poly',
            'rbf',
            'sigmoid'
        ])
        vbox.addWidget(self.type_combo)

        self.penaltyc = PLineEdit(1.0, parse=float)
        self.gamma = PLineEdit('auto', parse=float)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('Penalty C:'),
                                self.penaltyc,
                                QtWidgets.QLabel('Gamma:'),
                                self.gamma,
                                stretch=[0, 1, 0, 1]))

        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        vbox.addWidget(HWidgets(None, self.btn_predict, stretch=[1, 0]))

    def on_predict_clicked(self):
        params = {
            'clf': 'svm',
            'kernel': self.type_combo.currentText(),
            'C': self.penaltyc.value(),
            'gamma': self.gamma.value()
        }

        self.predict.emit(params)


class OnlineWidget(QtWidgets.QWidget):
    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(OnlineWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Loss function:', [
            # 'hinge',
            'log',
            'modified_huber',
            # 'squared_hinge',
            # 'perceptron'
        ])
        vbox.addWidget(self.type_combo)

        self.penalty = TComboBox('Regularization:', [
            'none',
            'l1',
            'l2',
            'elasticnet',
        ])
        vbox.addWidget(self.penalty)

        self.alpha = PLineEdit(0.0001, parse=float)
        self.n_iter = PLineEdit(5, parse=int)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('alpha:'),
                                self.alpha,
                                QtWidgets.QLabel('Num Iter:'),
                                self.n_iter,
                                stretch=[0, 1, 0, 1]))

        self.chunks = PLineEdit(None, parse=int)

        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        vbox.addWidget(HWidgets(QtWidgets.QLabel('Chunks:'),
                                self.chunks,
                                None, self.btn_predict,
                                stretch=[0, 0, 1, 0]))

    def on_predict_clicked(self):
        params = {
            'clf': 'sgd',
            'loss': self.type_combo.currentText(),
            'penalty': self.penalty.currentText(),
            'alpha': self.alpha.value(),
            'n_iter': self.n_iter.value(),
            'chunks': self.chunks.value()
        }

        self.predict.emit(params)


class SaveClassifier(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(SaveClassifier, self).__init__(parent=parent)

        self.DM = DataModel.instance()
        self.LBLM = LabelManager.instance()
        self.selected_level = -1

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self.btn_save_clf = ActionButton('Save Classifier to Disk')
        self.btn_save_clf.clicked.connect(self.on_save_clf_clicked)
        vbox.addWidget(HWidgets(None, None, self.btn_save_clf, None,
                                stretch=[0, 0, 1, 0]))
        self.DM.level_predicted.connect(self.on_level_predicted)

    def on_level_predicted(self, idx):
        self.select_level(idx)

    def select_level(self, idx):
        self.selected_level = idx

    def on_save_clf_clicked(self):
        if not self.DM.has_classifier():
            QtWidgets.QMessageBox.critical(self, "Error", "A classifier has not been created yet!")
            return
        else:
            attrs = dict(levelid=self.selected_level, label=self.LBLM.idxs(self.selected_level),
                         names=self.LBLM.names(self.selected_level), colors=self.LBLM.colors(self.selected_level),
                         visible=list(map(int, self.LBLM.visibility(self.selected_level))),
                         parent_levels=self.LBLM.parent_levels(self.selected_level),
                         parent_labels=self.LBLM.parent_labels(self.selected_level),
                         parent_colours=[self.LBLM.get(level, label).color for level, label in
                                         zip(self.LBLM.parent_levels(self.selected_level),
                                             self.LBLM.parent_labels(self.selected_level))])

            root_dir = self.DM.wspath
            output_dir = op.join(root_dir, "classifiers")
            os.makedirs(output_dir, exist_ok=True)
            filename = op.join(output_dir, "classifier.h5")
            filter = "Classifier (*.h5)"
            path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save Classifier', filename, filter)
            if path is not None and len(path) > 0:
                log.info('+ Saving classifier to {}'.format(path))
                self.DM.save_classifier(path, attrs)


class TrainPredict(PredWidgetBase):
    def __init__(self):
        super(TrainPredict, self).__init__(parent=None)

        self.vbox.addWidget(SubHeaderLabel('Classifier'))

        self.train_alg = TComboBox('Classifier type:', [
            'Ensemble',
            'SVM',
            'Online Linear'
        ])
        self.train_alg.currentIndexChanged.connect(self.on_classifier_changed)
        self.vbox.addWidget(self.train_alg)

        self.clf_container = QtWidgets.QWidget()
        vbox2 = QtWidgets.QVBoxLayout()
        vbox2.setContentsMargins(0, 0, 0, 0)
        self.clf_container.setLayout(vbox2)

        self.ensembles = EnsembleWidget()
        self.ensembles.train_predict.connect(self.on_predict)
        self.svm = SVMWidget()
        self.svm.predict.connect(self.on_predict)
        self.online = OnlineWidget()
        self.online.predict.connect(self.on_predict)

        self.clf_container.layout().addWidget(self.ensembles)
        self.vbox.addWidget(self.clf_container)

    def on_classifier_changed(self, idx):
        if idx == 0:
            self.clf_container.layout().addWidget(self.ensembles)
            self.svm.setParent(None)
            self.online.setParent(None)
        elif idx == 1:
            self.clf_container.layout().addWidget(self.svm)
            self.ensembles.setParent(None)
            self.online.setParent(None)
        else:
            self.clf_container.layout().addWidget(self.online)
            self.svm.setParent(None)
            self.ensembles.setParent(None)

    def run_prediction(self, y_data, p_data, level_params, desc_params, ref_params,
                       clf_params, out_labels, out_confidence, level):
        """
        Overrides method from parent class
        """

        self.launcher.run(ac.predict_proba, y_data=y_data, p_data=p_data, train=True,
                          level_params=level_params, desc_params=desc_params,
                          clf_params=clf_params, ref_params=ref_params,
                          out_labels=out_labels, out_confidence=out_confidence,
                          cb=self.on_predicted,
                          caption='Predicting labels for Level {}'.format(level))


class UncertainLabelWidget(QtWidgets.QWidget):
    save = QtCore.pyqtSignal(int)

    def __init__(self, idx, name, color, parent=None):
        super(UncertainLabelWidget, self).__init__(parent=parent)

        self.idx = idx

        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        self.name = BLabel(str(name))
        self.name.setMinimumWidth(120)
        hbox.addWidget(self.name)

        self.color = ColorButton(color, clickable=False)
        hbox.addWidget(self.color, 1)

        self.apply = ActionButton('Save')
        hbox.addWidget(self.apply)

        self.apply.clicked.connect(self.on_save_label)

    def on_save_label(self):
        self.save.emit(self.idx)


class Uncertainty(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(Uncertainty, self).__init__(parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()

        self.selected_level = -1
        self.parent_level = None
        self.parent_label = -1

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self.level_header = SubHeaderLabel('No Level Predicted Yet')
        vbox.addWidget(self.level_header)

        self.combo_viz = TComboBox('Visualization:', ['Predictions', 'Confidence'])
        vbox.addWidget(self.combo_viz)

        self.labels = {}

        self.slider_thresh = CSlider('none', current=0)
        vbox.addWidget(HWidgets('Confidence:', self.slider_thresh, stretch=[0, 1]))

        self.txt_from = HSize3D('From', default=(0, 0, 0), txtwidth=50)
        self.txt_to = HSize3D('To', default=self.DM.data_shape, txtwidth=50)
        vbox.addWidget(self.txt_from)
        vbox.addWidget(self.txt_to)

        self.combo_viz.currentIndexChanged.connect(self.on_viz_changed)
        self.slider_thresh.valueChanged.connect(self.on_thresh_changed)
        self.DM.level_predicted.connect(self.on_level_predicted)

    def on_level_predicted(self, idx):
        self.select_level(idx)

    def on_viz_changed(self, idx):
        if not self.LM.isin('Predictions', 'Predictions'):
            return

        labels_name = 'predictions/predictions'
        confidence_name = 'predictions/confidence'
        if idx == 0 and self.DM.attr(labels_name, 'active') == True:
            data = labels_name
            labels = self.DM.attr(labels_name, 'labels')
            colors = self.LBLM.colors(self.selected_level)
            idxs = self.LBLM.idxs(self.selected_level)
            maxid = 0 if len(idxs) == 0 else max(idxs) + 1
            cmaplist = ['#000000'] * maxid
            for idx, color in zip(idxs, colors):
                cmaplist[idx] = color
            cmap = ListedColormap(cmaplist)
        else:
            data = confidence_name
            cmap = 'viridis'
            maxid = 1

        layer = self.LM.get('Predictions', 'Predictions')
        layer.data = data
        layer.cmap = cmap
        layer.vmin = 0
        layer.vmax = maxid

        if self.combo_viz.currentIndex() == 1:
            layer.threshold = None
        else:
            layer.threshold = (confidence_name, self.slider_thresh.value() / 100.)

        self.LM.setVisible('Predictions', 'Predictions', True)
        self.LM.update()

    def on_thresh_changed(self, thresh):
        self.on_viz_changed(self.combo_viz.currentIndex())

    def select_level(self, level):
        self.level_header.setText('Confidence Tool for Level {}'.format(level))
        self.selected_level = level

        for labelobj in self.labels.values():
            labelobj.setParent(None)

        self.labels = {}

        if level >= 0:
            for label in self.LBLM.labels(level):
                labelobj = UncertainLabelWidget(label.idx, label.name, label.color)
                labelobj.save.connect(self.on_save_label)
                self.layout().addWidget(labelobj)
                self.labels[label.idx] = labelobj

        self.on_viz_changed(self.combo_viz.currentIndex())

    def on_save_label(self, idx):
        if not self.LM.isin('Predictions', 'Predictions'):
            QtWidgets.QMessageBox.critical(self, 'Error', 'No label predicted')
            return

        f = self.txt_from.value()
        t = self.txt_to.value()
        s1 = slice(f[0], t[0])
        s2 = slice(f[1], t[1])
        s3 = slice(f[2], t[2])

        labels_name = 'predictions/predictions'
        uncertainty_name = 'predictions/confidence'
        threshold = self.slider_thresh.value() / 100.

        launcher = Launcher.instance()
        launcher.setup('Saving predictions for [{}]'.format(self.selected_level))
        log.info('+ Loading data into memory')
        pred = self.DM.load_slices(labels_name)
        conf = self.DM.load_slices(uncertainty_name)

        log.info('+ Masking annotations')
        mask = np.zeros(pred.shape, dtype=np.bool)
        np.logical_and((pred[s1, s2, s3] == idx), (conf[s1, s2, s3] >= threshold),
                       out=mask[s1, s2, s3])

        log.info('+ Saving results to disk')
        dataset = self.LBLM.dataset(self.selected_level)
        target = self.DM.load_slices(dataset)
        target[mask] = idx
        self.DM.write_slices(dataset, target)
        self.LM.update()

        launcher.cleanup()


class Training(Plugin):
    name = 'Train classifier'

    def __init__(self, parent=None):
        super(Training, self).__init__(ptype=Plugin.Plugin, parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        # Levels
        dummy = QtWidgets.QWidget()
        vbox = QtWidgets.QVBoxLayout()
        dummy.setLayout(vbox)
        self.use_level = TComboBox('Predict Level:', ['Level {}'.format(i) for i in self.LBLM.levels()])
        self.use_parent = TComboBox('Constrain Region:', ['None'])

        self.use_level.currentIndexChanged.connect(self.on_level_selected)
        self.use_parent.currentIndexChanged.connect(self.on_parent_selected)

        vbox.addWidget(self.use_level)
        vbox.addWidget(self.use_parent)

        self.LBLM.levelAdded.connect(self.on_levels_changed)
        self.LBLM.levelLoaded.connect(self.on_levels_changed)
        self.LBLM.levelRemoved.connect(self.on_levels_changed)

        self.LBLM.labelAdded.connect(self.on_label_added)
        self.LBLM.labelLoaded.connect(self.on_label_added)
        self.LBLM.labelNameChanged.connect(self.on_label_name_changed)
        self.LBLM.labelRemoved.connect(self.on_label_removed)

        self.DM.level_predicted.connect(self.on_level_predicted)

        self.addWidget(dummy)
        self.addWidget(HeaderLabel('Train New Classifier'))
        self.train_widget = TrainPredict()
        self.addWidget(self.train_widget)

        self.addWidget(HeaderLabel('Update annotations'))
        self.uncertainty_widget = Uncertainty()
        self.addWidget(self.uncertainty_widget)

        self.parent_labels = []
        self.levels = self.LBLM.levels()

        self.addWidget(HeaderLabel('Save Trained Classifier'))
        self.save_classifier_widget = SaveClassifier()
        self.addWidget(self.save_classifier_widget)
        if not self.DM.has_classifier():
            self.save_classifier_widget.btn_save_clf.setEnabled(False)

    def on_level_selected(self, idx):
        if idx < 0:
            self.train_widget.select_level(None, None, -1)
            return
        self.train_widget.select_level(list(self.levels)[idx], None, -1)

        self.parent_labels = [(None, None)]
        self.use_parent.clear()
        self.use_parent.addItem('None')
        for i in range(len(self.levels)):
            if i == idx:
                continue
            prev_lvl = self.use_level.itemText(i)
            for label in self.LBLM.labels(self.levels[i]):
                self.parent_labels += [(self.levels[i], label)]
                self.use_parent.addItem('{}/{}'.format(prev_lvl, label.name))

    def on_parent_selected(self, idx):
        if idx > 0:
            self.train_widget.parent_level = self.parent_labels[idx][0]
            self.train_widget.parent_label = self.parent_labels[idx][1].idx
        else:
            self.train_widget.parent_level = None
            self.train_widget.parent_label = -1

    def on_levels_changed(self):
        self.levels = self.LBLM.levels()
        self.use_level.updateItems(['Level {}'.format(i) for i in self.levels])
        self.parent_level = None

    def on_label_added(self, level, dataset, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            self.use_level.setCurrentIndex(self.levels.index(level))

    def on_label_name_changed(self, level, ds, label, name):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                prev_lvl = self.use_level.itemText(self.levels.index(lvl))
                self.use_parent.setItemText(idx, '{}/{}'.format(prev_lvl, name))

    def on_label_removed(self, level, ds, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                if self.train_widget.parent_level == level and \
                                self.train_widget.parent_label == label:
                    self.use_parent.setCurrentIndex(0)
                del self.parent_labels[idx]
                self.use_parent.removeItem(idx)

    def on_level_predicted(self, level):
        self.save_classifier_widget.btn_save_clf.setEnabled(True)



from .base import Plugin
from .visualization import Visualization
from .roi import ROI
from .superregions import SuperRegions
from .channels import FeatureChannels
from .training import Training
from .annotations import Annotations
from .export import Export
from .load_classifier import PretrainedClassifier




import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets
import os.path as op
import os
import logging as log
from ..core import Singleton

from matplotlib.colors import ListedColormap

from .base import Plugin, PredWidgetBase
from ..widgets import HeaderLabel, TComboBox
from .training import Uncertainty, UncertainLabelWidget
from ..core import DataModel, LayerManager, Launcher, LabelManager
from .. import actions as ac
from ..widgets import ActionButton
from ..lib._features import find_boundaries




class PredictButtonWidget(QtWidgets.QWidget):
    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(PredictButtonWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vbox)
        self.DM = DataModel.instance()
        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        vbox.addWidget(self.btn_predict)

    def on_predict_clicked(self):
        # Get classifier information from file via DM
        params = self.DM.get_classifier_params()
        self.predict.emit(params)


class Predict(PredWidgetBase):
    save_signal = QtCore.pyqtSignal()

    def __init__(self):
        super(Predict, self).__init__(parent=None)

        self.predict_btn_widget = PredictButtonWidget()
        self.predict_btn_widget.predict.connect(self.on_predict)
        self.vbox.addWidget(self.predict_btn_widget)

        # Stores whether launcher post() method is connected to classifier save function
        self.sv_clf_conn = False


    def run_prediction(self, y_data, p_data, level_params, desc_params, ref_params,
                       clf_params, out_labels, out_confidence, level):
        """
        Overrides method from parent class
        """
        append = False
        train = False

        # Test for new annotations
        if self.DM.new_annotations_added(y_data):
            # 1. Ask if they would like to append new data to existing and train a new classifier
            answer = QtWidgets.QMessageBox.question(self,
                                           "Append New Data?",
                                           "Additional annotation data has been added, would you"
                                           "\nlike to train a new classifier with this data included?"
                                           "\n\nIf you chose 'No', the prediction will be repeated without\nthis new data.",
                                           QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

            if answer == QtWidgets.QMessageBox.Yes:
                # Add a signal here to enable "Save new classifier" button
                self.launcher.post.connect(self.enable_save_classifer_btn)
                self.sv_clf_conn = True
                # Append and train and predict with new classifier
                append = True
                train = True

        self.launcher.run(ac.predict_proba, y_data=y_data, p_data=p_data, append=append,
                          train=train, level_params=level_params, desc_params=desc_params,
                          clf_params=clf_params, ref_params=ref_params,
                          out_labels=out_labels, out_confidence=out_confidence,
                          cb=self.on_predicted,
                          caption='Predicting labels for Level {}'.format(level))

    def enable_save_classifer_btn(self):

        self.save_signal.emit()

        if self.sv_clf_conn:
            self.launcher.post.disconnect(self.enable_save_classifer_btn)
            self.sv_clf_conn = False


class LoadClassifier(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(LoadClassifier, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()

        self.load = ActionButton('Load Trained Classifier')
        vbox.addWidget(self.load)
        vbox.addStretch(1)
        self.setLayout(vbox)

class ClassifierInfo(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(ClassifierInfo, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()

        self.info_label = QtWidgets.QLabel()
        self.info_label.setAlignment(QtCore.Qt.AlignCenter)
        my_font = QtGui.QFont()
        my_font.setBold(True)
        self.info_label.setFont(my_font)
        self.info_label.hide()
        vbox.addWidget(self.info_label)
        vbox.addStretch(1)
        self.setLayout(vbox)

    def set_text(self, text):
        self.info_label.setText(text)
        self.info_label.show()

class ApplySettings(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(ApplySettings, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()

        self.apply = ActionButton('Calculate Channels && Supervoxels')
        self.apply.setEnabled(False)
        vbox.addWidget(self.apply)
        vbox.addStretch(1)
        self.setLayout(vbox)


class SaveClassifier(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(SaveClassifier, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()

        self.save = ActionButton('Save New Classifier')
        self.save.setEnabled(False)
        vbox.addWidget(self.save)
        vbox.addStretch(1)
        self.setLayout(vbox)


@Singleton
class PretrainedClassifier(Plugin):

    name = 'Pretrained classifier'

    def __init__(self, parent=None):
        super().__init__(ptype=Plugin.Plugin, parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        # Stores whether launcher post() method is connected to supervoxel calculation function
        self.svx_conn = False

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        loadLabel = HeaderLabel('Use Pre-Trained Classifier')
        vbox.addWidget(loadLabel)
        self.load_classifier =  LoadClassifier()
        vbox.addWidget(self.load_classifier)
        self.load_classifier.load.clicked.connect(self.on_load_classifier)

        self.classifier_info = ClassifierInfo()
        vbox.addWidget(self.classifier_info)

        self.apply_settings = ApplySettings()
        vbox.addWidget(self.apply_settings)
        self.apply_settings.apply.clicked.connect(self.on_apply_settings)

        # Levels
        self.use_level = TComboBox('Predict Level:', ['Level {}'.format(i) for i in self.LBLM.levels()])
        self.use_parent = TComboBox('Constrain Region:', ['None'])

        self.use_level.currentIndexChanged.connect(self.on_level_selected)
        self.use_parent.currentIndexChanged.connect(self.on_parent_selected)

        vbox.addWidget(self.use_level)
        vbox.addWidget(self.use_parent)

        self.LBLM.levelAdded.connect(self.on_levels_changed)
        self.LBLM.levelLoaded.connect(self.on_levels_changed)
        self.LBLM.levelRemoved.connect(self.on_levels_changed)

        self.LBLM.labelAdded.connect(self.on_label_added)
        self.LBLM.labelLoaded.connect(self.on_label_added)
        self.LBLM.labelNameChanged.connect(self.on_label_name_changed)
        self.LBLM.labelRemoved.connect(self.on_label_removed)
        self.predict_widget = Predict()
        vbox.addWidget(self.predict_widget)
        if not self.DM.has_classifier():
            self.predict_widget.predict_btn_widget.btn_predict.setEnabled(False)
        self.predict_widget.save_signal.connect(self.enable_save_button)

        vbox.addWidget(HeaderLabel('Update annotations'))
        self.uncertainty_widget = Uncertainty()
        vbox.addWidget(self.uncertainty_widget)
        vbox.addStretch(1)

        vbox.addWidget(HeaderLabel('Save Modified Classifier'))
        self.save_clf_widget = SaveClassifier()
        self.save_clf_widget.save.clicked.connect(self.on_save_clf_clicked)
        vbox.addWidget(self.save_clf_widget)
        vbox.addStretch(1)

    def enable_save_button(self):
        self.save_clf_widget.save.setEnabled(True)

    def on_save_clf_clicked(self):
        if not self.DM.has_classifier():
            QtWidgets.QMessageBox.critical(self, "Error", "A classifier has not been created yet!")
            return
        else:
            current_level = self.predict_widget.selected_level
            attrs = dict(levelid=current_level,
                         label=self.LBLM.idxs(current_level),
                         names=self.LBLM.names(current_level),
                         colors=self.LBLM.colors(current_level),
                         visible=list(map(int, self.LBLM.visibility(current_level))),
                         parent_levels=self.LBLM.parent_levels(current_level),
                         parent_labels=self.LBLM.parent_labels(current_level),
                         parent_colours=[self.LBLM.get(level, label).color for level, label in
                                         zip(self.LBLM.parent_levels(current_level),
                                             self.LBLM.parent_labels(current_level))])

            root_dir = self.DM.wspath
            output_dir = op.join(root_dir, "classifiers")
            os.makedirs(output_dir, exist_ok=True)
            filename = op.join(output_dir, "modified_classifier.h5")
            filter = "Classifier (*.h5)"
            path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save Classifier', filename, filter)
            if path is not None and len(path) > 0:
                log.info('+ Saving classifier to {}'.format(path))
                self.DM.save_classifier(path, attrs)
                # Wipe the latest set of annotations to allow for more to be added
                self.LBLM.removeLevel(current_level, force=True)
                self.load_label_info()

    def on_load_classifier(self):
        root_dir = self.DM.wspath
        input_dir = op.join(root_dir, "classifiers")
        filter = "Classifier (*.h5)"
        path, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Load Classifier', input_dir, filter)

        success = None
        if path is not None and len(path) > 0:
            success = self.DM.load_classifier(path)
        if success:
            self.disable_train_tab()
            self.classifier_info.set_text(self.DM.clf_saved_name)
            # Get a dictionary with all the metadata settings
            self.meta_data_result = self.DM.load_saved_settings_from_file(path, has_classifier=True)

            ##### Labels #####
            self.load_label_info()
            level_list = self.LBLM.levels()
            label_list = [self.LBLM.labels(level) for level in level_list]
            label_names = []
            for label_sublist in label_list:
                for label in label_sublist:
                    label_names.append(label.name)

            self.launcher.info.emit("{} classifier loaded."
                                    "\nLevel {} loaded:\n{}".format(self.DM.clf_name, level_list, label_names))
            self.apply_settings.apply.setEnabled(True)

        else:
            self.launcher.error.emit("No classifier was loaded. Not applying settings")

    def disable_train_tab(self):
        plugins = self.parent().findChildren(Plugin)
        for plugin in plugins:
            if plugin.name == "Train classifier":
                plugin.setEnabled(False)

    def load_label_info(self):
        params = self.meta_data_result['lbl_attrs']
        levelid = params['levelid']
        # First create an empty dataset
        dataset = 'annotations/annotations{}'.format(levelid)
        self.DM.create_empty_dataset(dataset, shape=self.DM.data_shape,
                                     dtype=np.int16, params=params,
                                     fillvalue=-1)
        self.LBLM.loadLevel(levelid, dataset)

    def on_apply_settings(self):
        # Will calculate supervoxels after feature channels calculated
        self.calculate_feature_channels()

    def calculate_feature_channels(self):
        """
        Calculates feature channels with settings retrieved from HDF5 file.
        Attaches signal to the launcher to trigger supervoxel calculation when done.
        """
        # Check whether feature channels already exist

        if len(self.DM.available_channels()) == 0:
            features = []
            channels = self.meta_data_result['channel_list']

            dialog_result = self.confirm_settings_compute()
            if dialog_result == QtWidgets.QMessageBox.Yes:
                append_later = []
                # For each feature, create an empty file and generate a dictionary of parameters
                for channel in channels:
                    if channel != 'data':
                        metadata = self.meta_data_result[channel]
                        log.info("+ Queueing {}".format(metadata['feature_name']))
                        initial_params = dict(active=False, feature_idx=metadata['feature_idx'],
                                              feature_type=metadata['feature_type'], feature_name=metadata['feature_name'])
                        if len(initial_params) > 0:
                            self.DM.create_empty_dataset(metadata['out'], shape=self.DM.data_shape,
                                                         dtype=np.float32, params=initial_params,
                                                         fillvalue=np.nan)
                        clamp = None
                        if metadata['clamp'] == True:
                            clamp = metadata['evmin'], metadata['evmax']

                        info_dict = {
                                'name': metadata['feature_name'],
                                'source': metadata['source'],
                                'clamp': clamp,
                                'params': metadata,
                                'out': metadata['out'],
                                'idx': metadata['feature_idx'],
                                'feature': metadata['feature_type']
                            }

                        if metadata['source'] != '/data': # Non standard source for the feature
                            # Search for source channel
                            found_dicts = [item for item in features if item['out'] == metadata['source']]
                            if found_dicts:
                                # It is found, carry on as before
                                features.append(info_dict)
                            else:
                                # The source isn't there, so put this feature to the back of the queue
                                append_later.append(info_dict)
                        else: # Standard source for the feature
                            features.append(info_dict)

                # Append the final dictionaries with non-standard data sources
                features += append_later

                # Connect a signal from the launcher to initiate the supervoxel calculation when done
                if self.meta_data_result.get('sv_attrs', None):
                    self.launcher.post.connect(self.calculate_supervoxels)
                    self.svx_conn = True
                # Compute the feature channels
                self.launcher.run(ac.compute_all_channel, features=features,
                                  caption='Computing Multiple Features',
                                  cb=self.on_channels_calculated)
            else:
                log.info("Generation of feature channels cancelled")
        else:
            self.launcher.info.emit('Feature channels already exist. Not calculating new ones.')
            if not self.DM.has_grp('supervoxels') and self.meta_data_result.get('sv_attrs', None):
                self.calculate_supervoxels()
            elif self.meta_data_result.get('sv_attrs', None):
                pass
            else:
                self.launcher.info.emit('Supervoxels already exist. Not calculating new ones.')
            if self.DM.has_classifier():
                self.predict_widget.predict_btn_widget.btn_predict.setEnabled(True)

    def confirm_settings_compute(self):
        """
        Helper function to display dialog box
        """
        channels = self.meta_data_result['channel_list']
        full_names = [self.meta_data_result[name]['feature_name']
                      for name in channels
                      if self.meta_data_result[name]['feature_name'] != "data"]
        supervox_params = self.meta_data_result.get('sv_attrs', None)
        supervox_string = "" if supervox_params is None else "\n\nThe following supervoxels will be" \
                                                             " calculated:\n\nShape: {}\nSpacing: {}\nCompactness:" \
                                                             " {}\nSource: {}".format(", ".join(map(str, supervox_params['sp_shape'])),
                                                             ", ".join(map(str, supervox_params['spacing'])),
                                                             supervox_params['compactness'],
                                                             supervox_params['source'])
        return QtWidgets.QMessageBox.question(self,
                                            "Confirm calculation",
                                            "The following feature channels will be calculated:\n\n{}{}"
                                            .format("\n".join(full_names), "" if not supervox_string else supervox_string),
                                            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

    def on_channels_calculated(self, results):
        """
        Once the channels have been calculated, add them to the Feature Channels tab
        :param results: 
        :return: 
        """
        feat_list = []
        for result in results:
            if result is None:
                continue
            out, idx, params = result
            self.DM.channel_computed.emit(out, params)
            fname = params['feature_name']
            ftype = params['feature_type']
            active = params['active']
            name = out.split('/')[1]
            keys = list(params.keys())
            keys.remove('feature_idx')
            keys.remove('feature_name')
            keys.remove('feature_type')
            keys.remove('active')
            channel_params = {k: params[k] for k in keys}
            log.info('* Channel {} {}'.format(fname, ftype))
            self.DM.clf_channel_computed.emit(idx, name, ftype, active, channel_params)
        # Special case 1 - Supervoxel Source Channel
            if not params.get('unique_sv_source', None):
                feat_list.append(fname)
        # Special case 2 - Data
        if 'data' in self.meta_data_result['channel_list']:
            feat_list.append('Data')
        self.predict_widget.use_desc.checkGivenItems(feat_list)

    def calculate_supervoxels(self):

        ##### Supervoxels #####

        sv_attrs = self.meta_data_result['sv_attrs']

        in_data = sv_attrs['source']
        self.sv_params = {
            'sp_shape': tuple(sv_attrs['sp_shape']),
            'spacing': tuple(sv_attrs['spacing']),
            'compactness': sv_attrs['compactness']
        }

        out_sv = 'supervoxels/supervoxels'
        out_svindex = 'supervoxels/supervoxels_idx'
        out_svtable = 'supervoxels/supervoxels_table'
        out_svedges = 'supervoxels/graph_edges'
        out_svweights = 'supervoxels/graph_edge_weights'

        self.DM.remove_dataset(out_sv)
        self.DM.remove_dataset(out_svindex)
        self.DM.remove_dataset(out_svtable)
        self.DM.remove_dataset(out_svedges)
        self.DM.remove_dataset(out_svweights)

        # Disconnect the signal from  the launcher to prevent a never-ending loop
        if self.svx_conn:
            self.launcher.post.disconnect(self.calculate_supervoxels)
            self.svx_conn = False
        Launcher.instance().run(ac.create_supervoxels,
                                dataset=in_data,
                                out_sv=out_sv,
                                out_svindex=out_svindex, out_svtable=out_svtable,
                                out_svedges=out_svedges, out_svweights=out_svweights,
                                cb=self.on_supervoxels,
                                caption='Creating SuperVoxels..',
                                **self.sv_params)

    def on_supervoxels(self, params):
        svlabels, svtotal, sortindex, sorttable, edges, weights = params
        self.update_supervoxel_layer(svlabels, sortindex, sorttable, svtotal, visible=True)
        if self.DM.has_classifier():
            self.predict_widget.predict_btn_widget.btn_predict.setEnabled(True)

    def update_supervoxel_layer(self, svlabels, sortindex, sorttable, total_sv, visible=False):
        if svlabels is not None:
            self.DM.svlabels = svlabels
            self.DM.svtotal = total_sv
            self.DM.svindex = sortindex
            self.DM.svtable = sorttable

        if self.LM.isin('SuperVoxels', 'Super-Regions'):
            self.LM.get('SuperVoxels', 'Super-Regions').data = self.DM.svlabels
            self.LM.setVisible('SuperVoxels', 'Super-Regions', True)
        else:
            cmap = ListedColormap(['MidnightBlue'])
            self.LM.addLayer(self.DM.svlabels, 'SuperVoxels', level='Super-Regions',
                             pre=find_boundaries, cmap=cmap, background=0,
                             vmin=0.1, vmax=1, visible=visible)
        self.LM.update()

    def on_level_selected(self, idx):
        if idx < 0:
            self.predict_widget.select_level(None, None, -1)
            return
        self.predict_widget.select_level(list(self.levels)[idx], None, -1)

        self.parent_labels = [(None, None)]
        self.use_parent.clear()
        self.use_parent.addItem('None')
        for i in range(len(self.levels)):
            if i == idx:
                continue
            prev_lvl = self.use_level.itemText(i)
            for label in self.LBLM.labels(self.levels[i]):
                self.parent_labels += [(self.levels[i], label)]
                self.use_parent.addItem('{}/{}'.format(prev_lvl, label.name))

    def on_parent_selected(self, idx):
        if idx > 0:
            self.predict_widget.parent_level = self.parent_labels[idx][0]
            self.predict_widget.parent_label = self.parent_labels[idx][1].idx
        else:
            self.predict_widget.parent_level = None
            self.predict_widget.parent_label = -1

    def on_levels_changed(self):
        self.levels = self.LBLM.levels()
        self.use_level.updateItems(['Level {}'.format(i) for i in self.levels])
        self.parent_level = None

    def on_label_added(self, level, dataset, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            self.use_level.setCurrentIndex(self.levels.index(level))

    def on_label_name_changed(self, level, ds, label, name):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                prev_lvl = self.use_level.itemText(self.levels.index(lvl))
                self.use_parent.setItemText(idx, '{}/{}'.format(prev_lvl, name))

    def on_label_removed(self, level, ds, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                if self.predict_widget.parent_level == level and \
                        self.predict_widget.parent_label == label:
                    self.use_parent.setCurrentIndex(0)
                del self.parent_labels[idx]
                self.use_parent.removeItem(idx)



#!/usr/bin/env python

import sys
import os
import logging
import numpy as np
import h5py as h5

import argparse
def main():
    survos_title = "SuRVoS: Super-Region Volume Segmentation workbench"

    parser = argparse.ArgumentParser(description=survos_title)
    parser.add_argument('--workspace', type=str, default=None, help='Load workspace')
    parser.add_argument('--gpu', type=int, default=-1, help='Select GPU')
    args = parser.parse_args()

    import warnings
    warnings.filterwarnings("ignore")

    import sip
    sip.setapi("QString", 2)

    import matplotlib as mpl
    mpl.use('Qt5Agg')
    logging.getLogger('matplotlib.font_manager').disabled = True

    from matplotlib import style
    style.use( u'seaborn-ticks')

    from survos.qt_compat import QtGui, QtCore, QtWidgets
    from survos.widgets import MainWindow

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    root.addHandler(ch)

    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QtWidgets.QApplication(sys.argv)
    app.setOrganizationName("DLS")
    app.setApplicationName("SuRVoS")

    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'survos.qcs')) as f:
        css = f.read()

    window = MainWindow(survos_title)
    window.setMinimumSize(1024, 768)
    window.setStyleSheet(css)
    window.showMaximized()


    if args.gpu >= 0:
        window.setGPU(args.gpu)

    if args.workspace is not None and os.path.isdir(args.workspace):
        window.load_workspace(args.workspace)

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()




import numpy as np
import pandas as pd
import seaborn as sns

from ..qt_compat import QtGui, QtCore, QtWidgets

from .. import actions as ac
from ..plugins import Plugin
from ..core import LabelManager, Launcher

from .mpl_widgets import MplCanvas
from .base import HeaderLabel, SComboBox, RoundedWidget, \
    TComboBox, HWidgets, SubHeaderLabel, ColorButton, RCheckBox


class CompareLevel(QtWidgets.QWidget):

    def __init__(self, name, parent=None):
        super(CompareLevel, self).__init__(parent=parent)

        self.LBLM = LabelManager.instance()

        vbox = QtWidgets.QVBoxLayout()
        vbox.setAlignment(QtCore.Qt.AlignTop)
        self.setLayout(vbox)
        vbox.addWidget(HeaderLabel('Level {}'.format(name)))

        self.level_combo = SComboBox()
        self.level_combo.addItem('Select level')
        self.level_combo.setMinimumWidth(200)
        vbox.addWidget(self.level_combo)

        self.container = QtWidgets.QWidget()
        vbox2 = QtWidgets.QVBoxLayout(self.container)
        vbox2.setAlignment(QtCore.Qt.AlignTop)
        vbox.addWidget(self.container, 1)

        self.LBLM.levelLoaded.connect(self.on_level_added)
        self.LBLM.levelAdded.connect(self.on_level_added)
        self.LBLM.levelRemoved.connect(self.on_level_removed)

        self.LBLM.labelAdded.connect(self.on_label_added)
        self.LBLM.labelRemoved.connect(self.on_label_removed)
        self.LBLM.labelNameChanged.connect(self.on_label_name)
        self.LBLM.labelColorChanged.connect(self.on_label_color)

        self.levels = [None]
        self.labels = {}
        self.level_selected = -1

        self.level_combo.currentIndexChanged.connect(self.on_level_changed)

    def on_level_added(self, level):
        self.levels.append(level)
        self.level_combo.addItem('Level {}'.format(level))
        if level == 0:
            self.on_level_changed(0)

    def on_level_removed(self, level):
        i = self.levels.index(level)
        self.level_combo.removeItem(i)
        del self.levels[i]

    def on_level_changed(self, level):
        for label in self.labels.values():
            label.setParent(None)
        self.labels.clear()

        self.level_selected = self.levels[level]
        try:
            for label in self.LBLM.labels(self.levels[level]):
                widget = CompareLabel(label.idx, label.name, label.color)
                self.labels[label.idx] = widget
                self.container.layout().addWidget(widget)
        except:
            pass

    def on_label_added(self, level, level_name, label, label_name):
        if level != self.level_selected or label in self.labels:
            return
        label_color = self.LBLM.get(level, label).color
        widget = CompareLabel(label, label_name, label_color)
        self.labels[label] = widget
        self.container.layout().addWidget(widget)

    def on_label_removed(self, level, level_name, label):
        if level != self.level_selected:
            return
        if label.idx in self.labels:
            self.labels[label.idx].setParent(None)
            del self.labels[label.idx]

    def on_label_name(self, level, level_name, label, label_name):
        if level != self.level_selected:
            return
        if label in self.labels:
            self.labels[label].setName(label_name)

    def on_label_color(self, level, level_name, label, label_color):
        if level != self.level_selected:
            return
        if label in self.labels:
            self.labels[label].setColor(label_color)

    def value(self):
        labels = {l.source: l.target for l in self.labels.values()}
        return dict(
            level=self.level_selected, labels=labels
        )


class CompareLabel(RoundedWidget):

    def __init__(self, idx, name, color, parent=None):
        super(CompareLabel, self).__init__(
            parent=parent, color=None, bg='#cde5e5', width=0)
        self.idx = idx
        self.name = name

        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(hbox)

        self.color = ColorButton(color=color, clickable=False)
        hbox.addWidget(self.color)

        self.label = QtWidgets.QLabel(name)
        hbox.addWidget(self.label, 1)

        self.combo = TComboBox('Target ID', ['Ignore'] + list(range(25)),
                               parse=int, default=-1)
        hbox.addWidget(self.combo)

    @property
    def source(self):
        return self.idx

    @property
    def target(self):
        return self.combo.value()

    def setName(self, name):
        self.label.setText(name)

    def setColor(self, color):
        self.color.setColor(color)


class ScoreWidget(RoundedWidget):

    def __init__(self, name, score, parent=None):
        super(ScoreWidget, self).__init__(
            parent=parent, color=None, bg='#cde5e5', width=0)
        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(hbox)
        slabel = QtWidgets.QLabel('{0:.4f}'.format(score))
        slabel.setFixedWidth(80)
        slabel.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        hbox.addWidget(HWidgets(name, None, slabel,
                                sctretch=[False, True, False]))


class ComparisonResult(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(ComparisonResult, self).__init__(parent=parent)
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        self.setMaximumHeight(400)
        vbox.addWidget(HeaderLabel('Results'))

        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)

        self.scores = QtWidgets.QWidget()
        self.scores.setMinimumWidth(300)
        vbox2 = QtWidgets.QVBoxLayout()
        vbox2.setAlignment(QtCore.Qt.AlignTop)
        self.scores.setLayout(vbox2)

        self.mplcanvas = MplCanvas(axisoff=False)

        hbox.addWidget(self.scores)
        hbox.addWidget(self.mplcanvas, 1)

    def setResults(self, results):
        scores = []
        if 'dice' in results:
            scores.append(('Dice Coefficient', results['dice']))
        if 'jacc' in results:
            scores.append(('Jaccard Index', results['jacc']))
        if 'cohen' in results:
            scores.append(('Cohen\'s Kappa', results['cohen']))

        countsA = results['countsA']
        countsB = results['countsB']
        countsAB = results['countsOverlap']
        indexes = results['indexes']

        #
        # SCORES
        #

        for i in reversed(range(self.scores.layout().count())):
            self.scores.layout().itemAt(i).widget().setParent(None)

        for name, score in scores:
            widget = ScoreWidget(name, score)
            self.scores.layout().addWidget(widget)

        #
        # COUNTS A
        #

        self.mplcanvas.fig.clear()
        self.mplcanvas.ax.clear()

        t = list(indexes - {-1})
        n = len(t)

        for i, idx in enumerate(t):
            x = range(3)
            y = [countsA[idx + 1], countsB[idx + 1], countsAB[idx + 1]]

            ax = self.mplcanvas.fig.add_subplot(1, n, i + 1)
            ax.bar(x, y, color=['#2aa22a', '#3399ff', '#ff5050'])
            ax.set_xticks([j + 0.35 for j in x])
            ax.set_xticklabels(['Level A', 'Level B', 'Overlap'])
            ax.set_title('Target {}'.format(idx))
            ax.set_ylabel('Volume area ([0,1] ratio)')

        sns.despine(self.mplcanvas.fig)
        self.mplcanvas.redraw()


class QuantitativeAnalysis(Plugin):
    name = 'Quantitative Analysis'

    def __init__(self, ptype=Plugin.Widget):
        super(QuantitativeAnalysis, self).__init__(ptype=ptype)

        self.LBLM = LabelManager.instance()

        self.layout.parent = None
        vbox = QtWidgets.QVBoxLayout(self)
        self.layout = vbox
        splitter = QtWidgets.QSplitter()

        self.level_widget1 = CompareLevel('A')
        scroll = QtWidgets.QScrollArea()
        scroll.setWidget(self.level_widget1)
        scroll.setWidgetResizable(True)
        splitter.addWidget(scroll)

        self.level_widget2 = CompareLevel('B')
        scroll = QtWidgets.QScrollArea()
        scroll.setWidget(self.level_widget2)
        scroll.setWidgetResizable(True)
        splitter.addWidget(scroll)

        self.results = ComparisonResult()

        vbox.addWidget(splitter, 1)
        self.dice = RCheckBox('Dice Coefficient')
        self.jacc = RCheckBox('Jaccard Score')
        self.jacc.setChecked(False)
        self.cohen = RCheckBox('Cohen\'s Kappa')
        self.cohen.setChecked(False)
        button = QtWidgets.QPushButton('Calculate Segmentation Coefficients')
        vbox.addWidget(HWidgets(None, self.dice, self.jacc,
                                self.cohen, button, None,
                                stretch=[True, False, False, False, False, False, True]))
        vbox.addWidget(self.results)

        button.clicked.connect(self.on_compare_clicked)

    def on_compare_clicked(self):
        l1 = self.level_widget1.value()
        l2 = self.level_widget2.value()

        launcher = Launcher.instance()

        if l1['level'] is None or l2['level'] is None:
            launcher.show_error('Level A or Level B is not selected')
            return

        levelA = self.LBLM.dataset(l1['level'])
        levelB = self.LBLM.dataset(l2['level'])
        labelsA = l1['labels']
        labelsB = l2['labels']

        launcher.run(ac.compare_segmentations, levelA=levelA, levelB=levelB,
                     labelsA=labelsA, labelsB=labelsB, cb=self.on_analysis,
                     dice=self.dice.isChecked(), jacc=self.jacc.isChecked(),
                     cohen=self.cohen.isChecked(),
                     caption='Comparing segmentations')

    def on_analysis(self, result):
        self.results.setResults(result)
